"""
Optimizers
==========

#. :class:`.NullOptimizer`
#. :class:`.MMFF`
#. :class:`.UFF`
#. :class:`.ETKDG`
#. :class:`.XTB`
#. :class:`.MacroModelForceField`
#. :class:`.MacroModelMD`
#. :class:`.MOPAC`
#. :class:`.OptimizerSequence`
#. :class:`.CageOptimizerSequence`
#. :class:`.TryCatchOptimizer`
#. :class:`.RaisingOptimizer`


Optimizers are objects used to optimize molecules. Each optimizer is
initialized with some settings and can optimize a molecule
with :meth:`~.Optimizer.optimize`.

.. code-block:: python

    import stk

    mol = stk.BuildingBlock('NCCCN', ['amine'])
    mmff = stk.MMFF()
    mmff.optimize(mol)

    # Optimizers also work with ConstructedMolecule objects.
    polymer = stk.ConstructedMolecule(
        building_blocks=[mol],
        topology_graph=stk.polymer.Linear('A', [0], n=3)
    )
    etkdg = stk.ETKDG()
    etkdg.optimize(polymer)

Sometimes it is desirable to chain multiple optimizations, one after
another. For example, before running an optimization, it may be
desirable to embed a molecule first, to generate an initial structure.
:class:`.OptimizerSequence` may be used for this.

.. code-block:: python

    # Create a new optimizer which chains the previously defined
    # mmff and etkdg optimizers.
    optimizer_sequence = stk.OptimizerSequence(etkdg, mmff)

    # Run each optimizer in sequence.
    optimizer_sequence.optimize(polymer)

By default, running :meth:`.Optimizer.optimize` twice on the same
molecule will perform an optimization a second time on a molecule. If
we want to skip optimizations on molecules which have already been
optimized we can use the `use_cache` flag.

.. code-block:: python

    caching_etkdg = stk.ETKDG(use_cache=True)
    # First optimize call runs an optimization.
    caching_etkdg.optimize(polymer)
    # Second call does nothing.
    caching_etkdg.optimize(polymer)

Caching is done on a per :class:`.Optimizer` basis. Just because the
molecule has been cached by one :class:`.Optimizer` instance does not
mean that a different :class:`.Optimizer` instance will no longer
optimize the molecule.

.. _`adding optimizers`:

Making New Optimizers
---------------------

New optimizers can be made by simply making a class which inherits the
:class:`.Optimizer` class. In addition to this, the new class must
define a :meth:`~.Optimizer.optimize` method. The method must take 1
mandatory `mol` parameter. :meth:`~.Optimizer.optimize` will take the
`mol` and change its structure in whatever way it likes. Beyond this
there are no requirements. New optimizers can be added into the
:mod:`.optimizers` submodule or into a new submodule.

"""

import logging
import numpy as np
import rdkit.Chem.AllChem as rdkit
import warnings
import os
from functools import wraps
import subprocess as sp
import uuid
import shutil
from ...utilities import (
    is_valid_xtb_solvent,
    XTBInvalidSolventError,
    XTBExtractor
)
import pywindow

logger = logging.getLogger(__name__)


def _add_cache_use(optimize):
    """
    Make :meth:`~Optimizer.optimize` use the :attr:`~Optimizer._cache`.

    Decorates `optimize` so that before running it checks if the
    :class:`.Molecule` has already been optimized by the
    optimizer. If so, and :attr:`~Optimizer.use_cache` is ``True``,
    then the molecule is skipped and no optimization is performed.

    Parameters
    ----------
    optimize : :class:`function`
        A function which is to have skipping added to it.

    Returns
    -------
    :class:`function`
        The decorated function.

    """

    @wraps(optimize)
    def inner(self, mol):
        if self._use_cache and mol in self._cache:
            logger.info(f'Skipping optimization on {mol}.')
        else:
            optimize(self, mol)
            if self._use_cache:
                self._cache.add(mol)

    return inner


class Optimizer:
    """
    A base class for optimizers.

    """

    def __init__(self, use_cache=False):
        """
        Initialize an :class:`Optimizer`.

        Parameters
        ----------
        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        # Holds every previously optimized molecule if use_cache is
        # true.
        self._cache = set()
        self._use_cache = use_cache

    def __init_subclass__(cls, **kwargs):
        cls.optimize = _add_cache_use(cls.optimize)
        return super().__init_subclass__(**kwargs)

    def is_caching(self):
        """
        ``True`` if the optimizer has caching turned on.

        Returns
        -------
        :class:`bool`
            ``True`` if the optimizer has caching turned on.

        """

        return self._use_cache

    def add_to_cache(self, mol):
        """
        Add a molecule to the cache.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be added to the cache.

        Returns
        -------
        None : :class:`NoneType`

        """

        self._cache.add(mol)

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        raise NotImplementedError()


class OptimizerSequence(Optimizer):
    """
    Applies optimizers in sequence.

    Examples
    --------
    Let's say we want to embed a molecule with ETKDG first and then
    minimize it with the MMFF force field.

    .. code-block:: python

        import stk

        mol = stk.BuildingBlock('NCCNCCN', ['amine'])
        optimizer = stk.OptimizerSequence(stk.ETKDG(), stk.MMFF())
        optimizer.optimize(mol)

    """

    def __init__(self, *optimizers, use_cache=False):
        """
        Initialize a :class:`OptimizerSequence` instance.

        Parameters
        ----------
        *optimizers : :class:`Optimizer`
            A number of optimizers, each of which gets applied to a
            molecule, based on the order given.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        self._optimizers = optimizers
        super().__init__(use_cache=use_cache)

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        for optimizer in self._optimizers:
            cls_name = optimizer.__class__.__name__
            logger.info(f'Using {cls_name} on "{mol}".')
            optimizer.optimize(mol)


class CageOptimizerSequence(Optimizer):
    """
    Applies :class:`Optimizer` objects to a cage.

    Before each :class:`Optimizer` in the sequence is applied to the
    cage, it is checked to see if it is collapsed. If it is
    collapsed, the optimization sequence ends immediately.

    Examples
    --------
    Let's say we want to embed a cage with ETKDG first and then
    minimize it with the MMFF force field.

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCNCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CC(C=O)C=O', ['aldehyde'])
        cage = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=stk.cage.FourPlusSix()
        )
        optimizer = stk.CageOptimizerSequence(stk.ETKDG(), stk.MMFF())
        optimizer.optimize(cage)

    """

    def __init__(self, *optimizers, use_cache=False):
        """
        Initialize a :class:`CageOptimizerSequence` instance.

        Parameters
        ----------
        *optimizers : :class:`Optimizer`
            The :class:`Optimizers` used in sequence to optimize
            cage molecules.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        self._optimizers = optimizers
        super().__init__(use_cache=use_cache)

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The cage to be optimized. It must have an attribute
            :attr:`num_windows`, which contains the expected number
            of windows if the cage is not collapsed.

        Returns
        -------
        None : :class:`NoneType`

        """

        for optimizer in self._optimizers:
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")
                loader = pywindow.molecular.Molecule.load_rdkit_mol
                pw_molecule = loader(mol.to_rdkit_mol())
                windows = pw_molecule.calculate_windows()
            logger.debug(f'Windows found: {windows}.')

            if windows is None or len(windows) != mol.num_windows:
                logger.info(f'"{mol}" is collapsed, exiting early.')
                return

            cls_name = optimizer.__class__.__name__
            logger.info(f'Using {cls_name} on "{mol}".')
            optimizer.optimize(mol)


class NullOptimizer(Optimizer):
    """
    Does not perform optimizations.

    """

    def optimize(self, mol):
        """
        Do not optimize `mol`.

        This function just returns immediately without changing the
        molecule.

        Parameters
        ----------
        mol : :class:`.Molecule`
            A molecule.

        Returns
        -------
        None : :class:`NoneType`

        """

        return


class TryCatchOptimizer(Optimizer):
    """
    Try to optimize with a :class:`Optimizer`, use another on failure.

    Examples
    --------
    .. code-block:: python

        import stk

        # Create some molecules to optimize.
        mol1 = stk.BuildingBlock('NCCN', ['amine'])
        mol2 = stk.BuildingBlock('CCCCC')
        mol3 = stk.BuildingBlock('O=CCCN')

        # Create an optimizer which may fail.
        uff = stk.UFF()

        # Create a backup optimizer.
        mmff = stk.MMFF()

        # Make an optimizer which tries to run raiser and if that
        # raises an error, will run mmff on the molecule instead.
        try_catch = stk.TryCatchOptimizer(
            try_optimizer=uff,
            catch_optimizer=mmff
        )

        # Optimize the molecules. In each case if the optimization with
        # UFF fails, MMFF is used to optimize the molecule instead.
        try_catch.optimize(mol1)
        try_catch.optimize(mol2)
        try_catch.optimzie(mol3)

    """

    def __init__(
        self,
        try_optimizer,
        catch_optimizer,
        use_cache=False
    ):
        """
        Initialize a :class:`TryCatchOptimizer` instance.

        Parameters
        ----------
        try_optimizer : :class:`Optimizer`
            The optimizer which is used initially to try and optimize a
            :class:`.Molecule`.

        catch_optimizer : :class:`Optimizer`
            If :attr:`try_optimizer` raises an error, this optimizer is
            run on the :class:`.Molecule` instead.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        self._try_optimizer = try_optimizer
        self._catch_optimizer = catch_optimizer
        super().__init__(use_cache=use_cache)

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        try:
            return self._try_optimizer.optimize(mol)
        except Exception:
            try_name = self._try_optimizer.__class__.__name__
            catch_name = self._catch_optimizer.__class__.__name__
            logger.error(
                f'{try_name} failed, trying {catch_name}.',
                exc_info=True
            )
            return self._catch_optimizer.optimize(mol)


class RaisingOptimizerError(Exception):
    ...


class RaisingOptimizer(Optimizer):
    """
    Raises and optimizes at random.

    This optimizer is used for debugging to simulate optimizations
    which sometimes complete successfully and sometimes
    randomly fail.

    Examples
    --------
    .. code-block:: python

        import stk

        mol = stk.BuildingBlock('NCCNCCN', ['amine'])
        partial_raiser = stk.RaisingOptimizer(
            optimizer=stk.ETKDG(),
            fail_chance=0.75
        )
        # 75 % chance an error will be raised by calling optimize.
        partial_raiser.optimize(mol)

    """

    def __init__(self, optimizer, fail_chance=0.5, use_cache=False):
        """
        Initialize :class:`PartialRaiser`.

        Parameters
        ----------
        optimizer : :class:`Optimizer`
            When the optimizer does not fail, it uses this
            :class:`Optimizer` to optimize molecules.

        fail_chance : :class:`float`, optional
            The probability that the optimizer will raise an error each
            time :meth:`optimize` is used.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        self._optimizer = optimizer
        self._fail_chance = fail_chance
        super().__init__(use_cache=use_cache)

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`RaisingOptimizerError`
            This error is raised at random.

        """

        if np.random.rand() < self._fail_chance:
            raise RaisingOptimizerError('Used RaisingOptimizer.')
        return self._optimizer.optimize(mol)


class MMFF(Optimizer):
    """
    Use the MMFF force field to optimize molecules.

    Examples
    --------
    .. code-block:: python

        import stk

        mol = stk.BuildingBlock('NCCNCCN', ['amine'])
        mmff = stk.MMFF()
        mmff.optimize(mol)

    """

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        rdkit_mol = mol.to_rdkit_mol()
        # Needs to be sanitized to get force field params.
        rdkit.SanitizeMol(rdkit_mol)
        rdkit.MMFFOptimizeMolecule(rdkit_mol)
        mol.update_from_rdkit_mol(rdkit_mol)


class UFF(Optimizer):
    """
    Use the UFF force field to optimize molecules.

    Examples
    --------
    .. code-block:: python

        import stk

        mol = stk.BuildingBlock('NCCNCCN', ['amine'])
        uff = stk.UFF()
        uff.optimize(mol)

    """

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        rdkit_mol = mol.to_rdkit_mol()
        # Needs to be sanitized to get force field params.
        rdkit.SanitizeMol(rdkit_mol)
        rdkit.UFFOptimizeMolecule(rdkit_mol)
        mol.update_from_rdkit_mol(rdkit_mol)


class ETKDG(Optimizer):
    """
    Uses the ETKDG [#]_ v2 algorithm to find an optimized structure.

    Examples
    --------
    .. code-block:: python

        import stk

        mol = stk.BuildingBlock('NCCNCCN', ['amine'])
        etkdg = stk.ETKDG()
        etkdg.optimize(mol)

    References
    ----------
    .. [#] http://pubs.acs.org/doi/pdf/10.1021/acs.jcim.5b00654

    """

    def __init__(self, random_seed=12, use_cache=False):
        """
        Initialize a :class:`ETKDG` instance.

        Parameters
        ----------
        random_seed : :class:`int`, optional
            The random seed to use.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        self._random_seed = random_seed
        super().__init__(use_cache=use_cache)

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        params = rdkit.ETKDGv2()
        params.clearConfs = True
        params.random_seed = self._random_seed

        rdkit_mol = mol.to_rdkit_mol()
        rdkit.EmbedMolecule(rdkit_mol, params)
        mol.set_position_matrix(
            position_matrix=rdkit_mol.GetConformer().GetPositions()
        )


class XTBOptimizerError(Exception):
    ...


class XTBConvergenceError(XTBOptimizerError):
    ...


class XTB(Optimizer):
    """
    Uses GFN-xTB [1]_ to optimize molecules.

    Notes
    -----
    When running :meth:`optimize`, this calculator changes the
    present working directory with :func:`os.chdir`. The original
    working directory will be restored even if an error is raised, so
    unless multi-threading is being used this implementation detail
    should not matter.

    If multi-threading is being used an error could occur if two
    different threads need to know about the current working directory
    as :class:`.XTB` can change it from under them.

    Note that this does not have any impact on multi-processing,
    which should always be safe.

    Furthermore, :meth:`optimize` will check that the
    structure is adequately optimized by checking for negative
    frequencies after a Hessian calculation. `max_runs` can be
    provided to the initializer to set the maximum number of
    optimizations which will be attempted at the given
    `opt_level` to obtain an optimized structure. However, we outline
    in the examples how to iterate over `opt_levels` to increase
    convergence criteria and hopefully obtain an optimized structure.
    The presence of negative frequencies can occur even when the
    optimization has converged based on the given `opt_level`.

    Attributes
    ----------
    incomplete : :class:`set` of :class:`.Molecule`
        A :class:`set` of molecules passed to :meth:`optimize` whose
        optimzation was incomplete.

    Examples
    --------
    Note that for :class:`.ConstructedMolecule` objects constructed by
    ``stk``, :class:`XTB` should usually be used in a
    :class:`.OptimizerSequence`. This is because xTB only uses
    xyz coordinates as input and so will not recognize the long bonds
    created during construction. An optimizer which can minimize
    these bonds should be used before :class:`XTB`.

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCNCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CCCC=O', ['aldehyde'])
        polymer = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=stk.polymer.Linear("AB", [0, 0], 3)
        )

        xtb = stk.OptimizerSequence(
            stk.UFF(),
            stk.XTB(xtb_path='/opt/gfnxtb/xtb', unlimited_memory=True)
        )
        xtb.optimize(polymer)

    By default, all optimizations with xTB are performed using the
    ``--ohess`` flag, which forces the calculation of a numerical
    Hessian, thermodynamic properties and vibrational frequencies.
    :meth:`optimize` will check that the structure is appropriately
    optimized (i.e. convergence is obtained and no negative vibrational
    frequencies are present) and continue optimizing a structure (up to
    `max_runs` times) until this is achieved. This loop, by
    default, will be performed at the same `opt_level`. The
    following example shows how a user may optimize structures with
    tigher convergence criteria (i.e. different `opt_level`)
    until the structure is sufficiently optimized. Furthermore, the
    calculation of the Hessian can be turned off using
    `max_runs` to ``1`` and `calculate_hessian` to ``False``.

    .. code-block:: python

        # Use crude optimization with max_runs=1 because this will
        # not achieve optimization and rerunning it is unproductive.
        xtb_crude = stk.XTB(
            xtb_path='/opt/gfnxtb/xtb',
            output_dir='xtb_crude',
            unlimited_memory=True,
            opt_level='crude',
            max_runs=1,
            calculate_hessian=True
        )
        # Use normal optimization with max_runs == 2.
        xtb_normal = stk.XTB(
            xtb_path='/opt/gfnxtb/xtb',
            output_dir='xtb_normal',
            unlimited_memory=True,
            opt_level='normal',
            max_runs=2
        )
        # Use vtight optimization with max_runs == 2, which should
        # achieve sufficient optimization.
        xtb_vtight = stk.XTB(
            xtb_path='/opt/gfnxtb/xtb',
            output_dir='xtb_vtight',
            unlimited_memory=True,
            opt_level='vtight',
            max_runs=2
        )

        optimizers = [xtb_crude, xtb_normal, xtb_vtight]
        for optimizer in optimizers:
            optimizer.optimize(polymer)
            if polymer not in optimizer.incomplete:
                break

    References
    ----------
    .. [1] https://xtb-docs.readthedocs.io/en/latest/setup.html

    """

    def __init__(
        self,
        xtb_path,
        gfn_version=2,
        output_dir=None,
        opt_level='normal',
        max_runs=2,
        calculate_hessian=True,
        num_cores=1,
        electronic_temperature=300,
        solvent=None,
        solvent_grid='normal',
        charge=0,
        num_unpaired_electrons=0,
        unlimited_memory=False,
        use_cache=False
    ):
        """
        Initialize a :class:`XTB` instance.

        Parameters
        ----------
        xtb_path : :class:`str`
            The path to the xTB executable.

        gfn_version : :class:`int`, optional
            Parameterization of GFN to use in xTB.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/basics.html.

        output_dir : :class:`str`, optional
            The name of the directory into which files generated during
            the optimization are written, if ``None`` then
            :func:`uuid.uuid4` is used.

        opt_level : :class:`str`, optional
            Optimization level to use.
            Can be one of ``'crude'``, ``'sloppy'``, ``'loose'``,
            ``'lax'``, ``'normal'``, ``'tight'``, ``'vtight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/optimization.html
            .

        max_runs : :class:`int`, optional
            Maximum number of optimizations to attempt in a row.

        calculate_hessian : :class:`bool`, optional
            Toggle calculation of the hessian and vibrational
            frequencies after optimization. ``True`` is required to
            check that the structure is completely optimized.
            ``False`` will drastically speed up the calculation but
            potentially provide incomplete optimizations and forces
            :attr:`max_runs` to be ``1``.

        num_cores : :class:`int`, optional
            The number of cores xTB should use.

        electronic_temperature : :class:`int`, optional
            Electronic temperature in Kelvin.

        solvent : :class:`str`, optional
            Solvent to use in GBSA implicit solvation method.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        solvent_grid : :class:`str`, optional
            Grid level to use in SASA calculations for GBSA implicit
            solvent.
            Can be one of ``'normal'``, ``'tight'``, ``'verytight'``
            or ``'extreme'``.
            For details see
            https://xtb-docs.readthedocs.io/en/latest/gbsa.html.

        charge : :class:`int`, optional
            Formal molecular charge.

        num_unpaired_electrons : :class:`int`, optional
            Number of unpaired electrons.

        unlimited_memory : :class: `bool`, optional
            If ``True`` :meth:`optimize` will be run without
            constraints on the stack size. If memory issues are
            encountered, this should be ``True``, however this may
            raise issues on clusters.

        use_cache : :class:`bool`, optional
            If ``True`` :meth:`optimize` will not run twice on the same
            molecule.

        """

        if solvent is not None:
            solvent = solvent.lower()
            if gfn_version == 0:
                raise XTBInvalidSolventError(
                    f'No solvent valid for version',
                    f' {gfn_version!r}.'
                )
            if not is_valid_xtb_solvent(gfn_version, solvent):
                raise XTBInvalidSolventError(
                    f'Solvent {solvent!r} is invalid for ',
                    f'version {gfn_version!r}.'
                )

        if not calculate_hessian and max_runs != 1:
            max_runs = 1
            logger.warning(
                'Requested that hessian calculation was skipped '
                'but the number of optimizations requested was '
                'greater than 1. The number of optimizations has been '
                'set to 1.'
            )

        self._xtb_path = xtb_path
        self._gfn_version = str(gfn_version)
        self._output_dir = output_dir
        self._opt_level = opt_level
        self._max_runs = max_runs
        self._calculate_hessian = calculate_hessian
        self._num_cores = str(num_cores)
        self._electronic_temperature = str(electronic_temperature)
        self._solvent = solvent
        self._solvent_grid = solvent_grid
        self._charge = str(charge)
        self._num_unpaired_electrons = str(num_unpaired_electrons)
        self._unlimited_memory = unlimited_memory
        self.incomplete = set()
        super().__init__(use_cache=use_cache)

    def _has_neg_frequencies(self, output_file):
        """
        Check for negative frequencies.

        Parameters
        ----------
        output_file : :class:`str`
            Name of output file with xTB results.

        Returns
        -------
        :class:`bool`
            Returns ``True`` if a negative frequency is present.

        """
        xtbext = XTBExtractor(output_file=output_file)
        # Check for one negative frequency, excluding the first
        # 6 frequencies.
        return any(x < 0 for x in xtbext.frequencies[6:])

    def _is_complete(self, output_file):
        """
        Check if xTB optimization has completed and converged.

        Parameters
        ----------
        output_file : :class:`str`
            Name of xTB output file.

        Returns
        -------
        :class:`bool`
            Returns ``False`` if a negative frequency is present.

        Raises
        -------
        :class:`XTBOptimizerError`
            If the optimization failed.

        :class:`XTBConvergenceError`
            If the optimization did not converge.

        """
        if output_file is None:
            # No simulation has been run.
            return False
        # If convergence is achieved, then .xtboptok should exist.
        if os.path.exists('.xtboptok'):
            # Check for negative frequencies in output file if the
            # hessian was calculated.
            # Return True if there exists at least one.
            if self._calculate_hessian:
                return not self._has_neg_frequencies(output_file)
            else:
                return True
        elif os.path.exists('NOT_CONVERGED'):
            raise XTBConvergenceError('Optimization not converged.')
        else:
            raise XTBOptimizerError('Optimization failed to complete')

    def _run_xtb(self, xyz, out_file):
        """
        Run GFN-xTB.

        Parameters
        ----------
        xyz : :class:`str`
            The name of the input structure ``.xyz`` file.

        out_file : :class:`str`
            The name of output file with xTB results.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Modify the memory limit.
        if self._unlimited_memory:
            memory = 'ulimit -s unlimited ;'
        else:
            memory = ''

        # Set optimization level and type.
        if self._calculate_hessian:
            # Do optimization and check hessian.
            optimization = f'--ohess {self._opt_level}'
        else:
            # Do optimization.
            optimization = f'--opt {self._opt_level}'

        if self._solvent is not None:
            solvent = f'--gbsa {self._solvent} {self._solvent_grid}'
        else:
            solvent = ''

        cmd = (
            f'{memory} {self._xtb_path} {xyz} '
            f'--gfn {self._gfn_version} '
            f'{optimization} --parallel {self._num_cores} '
            f'--etemp {self._electronic_temperature} '
            f'{solvent} --chrg {self._charge} '
            f'--uhf {self._num_unpaired_electrons}'
        )

        with open(out_file, 'w') as f:
            # Note that sp.call will hold the program until completion
            # of the calculation.
            sp.call(
                cmd,
                stdin=sp.PIPE,
                stdout=f,
                stderr=sp.PIPE,
                # Shell is required to run complex arguments.
                shell=True
            )

    def _run_optimizations(self, mol):
        """
        Run loop of optimizations on `mol` using xTB.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        :class:`bool`
            Returns ``True`` if the calculation is complete and
            ``False`` if the calculation is incomplete.

        """
        for run in range(self._max_runs):
            xyz = f'input_structure_{run+1}.xyz'
            out_file = f'optimization_{run+1}.output'
            mol.write(xyz)
            self._run_xtb(xyz=xyz, out_file=out_file)
            # Check if the optimization is complete.
            coord_file = 'xtbhess.coord'
            coord_exists = os.path.exists(coord_file)
            output_xyz = 'xtbopt.xyz'
            opt_complete = self._is_complete(out_file)
            if not opt_complete:
                if coord_exists:
                    # The calculation is incomplete.
                    # Update mol from xtbhess.coord and continue.
                    mol.update_from_file(coord_file)
                else:
                    # Update mol from xtbopt.xyz.
                    mol.update_from_file(output_xyz)
                    # If the negative frequencies are small, then GFN
                    # may not produce the restart file. If that is the
                    # case, exit optimization loop and warn.
                    self.incomplete.add(mol)
                    logging.warning(
                        f'Small negative frequencies present in {mol}.'
                    )
                    return False
            else:
                # Optimization is complete.
                # Update mol from xtbopt.xyz.
                mol.update_from_file(output_xyz)
                break
        return opt_complete

    def optimize(self, mol):
        """
        Optimize `mol`.

        Parameters
        ----------
        mol : :class:`.Molecule`
            The molecule to be optimized.

        Returns
        -------
        None : :class:`NoneType`

        """

        # Remove mol from self.incomplete if present.
        if mol in self.incomplete:
            self.incomplete.remove(mol)

        if self._output_dir is None:
            output_dir = str(uuid.uuid4().int)
        else:
            output_dir = self._output_dir
        output_dir = os.path.abspath(output_dir)

        if os.path.exists(output_dir):
            shutil.rmtree(output_dir)

        os.mkdir(output_dir)
        init_dir = os.getcwd()
        os.chdir(output_dir)

        try:
            complete = self._run_optimizations(mol)
        finally:
            os.chdir(init_dir)

        if not complete:
            self.incomplete.add(mol)
            logging.warning(f'Optimization is incomplete for {mol}.')



"""
Rotaxane
========

#. :class:`.NRotaxane`

"""


import numpy as np
import rdkit.Chem.AllChem as rdkit

from .topology_graph import TopologyGraph, VertexData, Vertex


class _AxleVertexData(VertexData):
    def get_vertex(self):
        return _AxleVertex(self)


class _AxleVertex(Vertex):
    def place_building_block(self, building_block, vertices, edges):
        building_block.set_centroid(self._position)
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        return {}


class _CycleVertexData(VertexData):
    """
    Holds data for a :class:`._CycleVertex`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. Must match the index in
        :attr:`TopologyGraph.vertices`.

    position : :class:`numpy.ndarray`
        The position of the vertex.

    edges : :class:`list` of :class:`.EdgeData`
        The edges connected to the vertex.

    cell : :class:`numpy.ndarray`
        The unit cell in which the vertex is found.

    flip : :class:`bool`
        If ``True`` any building block placed by the vertex will
        have its orientation along the chain flipped.

    """

    def __init__(self, x, y, z, flip):
        """
        Initialize a :Class:`._CycleVertexData` instance.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

        flip : :class:`bool`
            If ``True`` any building block placed by the vertex will
            have its orientation along the chain flipped.

        """

        self.flip = flip
        super().__init__(x, y, z)

    def clone(self, clear_edges=False):
        clone = super().clone(clear_edges)
        clone.flip = self.flip
        return clone

    def get_vertex(self):
        return _CycleVertex(self)


class _CycleVertex(Vertex):
    """
    Places the cycles in a :class:`NRotaxane`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def __init__(self, data):
        self._flip = data.flip
        super().__init__(data)

    def clone(self, clear_edges=False):
        clone = super().clone(clear_edges)
        clone._flip = self._flip
        return clone

    def place_building_block(self, building_block, vertices, edges):
        rdkit_mol = building_block.to_rdkit_mol()
        macrocycle = max(rdkit.GetSymmSSSR(rdkit_mol), key=len)
        cycle_normal = building_block.get_plane_normal(macrocycle)
        building_block.set_centroid(
            position=self._position,
            atom_ids=macrocycle
        )
        building_block.apply_rotation_between_vectors(
            start=cycle_normal,
            target=[-1 if self._flip else 1, 0, 0],
            origin=self._position
        )
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        return {}

    def __str__(self):
        return (
            f'Vertex(id={self.id}, '
            f'position={self._position.tolist()}, '
            f'flip={self._flip})'
        )


class NRotaxane(TopologyGraph):
    """
    Represents [n]rotaxane topology graphs.

    This class assumes one axle with (n-1) macrocycles threaded on it.
    The macrocycles are spaced evenly along the thread in repeating
    patterns. The threaded macrocycles can be described analagously
    to monomers in linear polymers, in terms of a repeating unit,
    except that no bonds are formed between them.

    The axle must be provided first to the `building_blocks` in
    :class:`.ConstructedMolecule.__init__`.

    Attributes
    ----------
    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    Examples
    --------
    .. code-block:: python

        import stk

        cycle = stk.ConstructedMolecule(
            building_blocks=[
                stk.BuildingBlock('[Br]CC[Br]', ['bromine'])
            ],
            topology_graph=stk.macrocycle.Macrocycle('A', 5)
        )
        axle = stk.ConstructedMolecule(
            building_blocks=[
                stk.BuildingBlock('NCCN', ['amine']),
                stk.BuildingBlock('O=CCC=O', ['aldehyde'])
            ],
            topology_graph=stk.polymer.Linear('AB', 7)
        )
        rotaxane = stk.ConstructedMolecule(
            building_blocks=[axle, cycle],
            topology_graph=stk.rotaxane.NRotaxane('A', 3)
        )

    The repeating unit can also be specified through the indices of
    the building blocks

    .. code-block:: python

        # r1 and r2 are different ways to write the same thing.
        r1 = stk.ConstructedMolecule(
            building_blocks=[axle, cycle, cycle2, cycle3],
            topology_graph=stk.rotaxane.NRotaxane('ACB', 3)
        )
        r2 = stk.ConstructedMolecule(
            building_blocks=[axle, cycle, cycle2, cycle3],
            topology_graph=stk.rotaxane.NRotaxane((1, 3, 2), 3)
        )

    :class:`.NRotaxane` shares many parameters with :class:`.Linear`,
    and the examples described there are also valid for this class.
    Be sure to read them.

    """

    def __init__(
        self,
        repeating_unit,
        num_repeating_units,
        orientations=None,
        random_seed=None,
        num_processes=1
    ):
        """
        Initialize a :class:`NRotaxane` instance.

        Parameters
        ----------
        repeating_unit : :class:`str` or :class:`tuple` of :class:`int`
            A string specifying the repeating unit of the macrocycles.
            For example, ``'AB'`` or ``'ABB'``. The first macrocycle
            passed to `building_blocks` is ``'A'`` and so on.

            The repeating unit can also be specified by the indices of
            `building_blocks`, for example ``'ABB'`` can be
            written as ``(1, 2, 2)``.

        num_repeating_units : :class:`int`
            The number of repeating units threaded along the axle.

        orientations : :class:`tuple` of :class:`float`, optional
            For each character in the repeating unit, a value
            between ``0`` and ``1`` (both inclusive) must be given in
            a :class:`tuple`. It indicates the probability that each
            macrocycle will have its orientation along the axle
            flipped. If ``0`` then the macrocycle is guaranteed not to
            flip. If ``1`` it is guaranteed to flip. This allows the
            user to create head-to-head or head-to-tail chains, as well
            as chain with a preference for head-to-head or head-to-tail
            if a number between ``0`` and ``1`` is chosen. If
            ``None`` then defaults to ``0`` in every case.

            It is also possible to supply an orientation for every
            cycle vertex in the final topology graph. In this case, the
            length of `orientations` must be equal to
            ``len(repeating_unit)*num_repeating_units``.

        random_seed : :class:`int`, optional
            The random seed to use when choosing random orientations.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        """

        if orientations is None:
            orientations = tuple(
                0. for i in range(len(repeating_unit))
            )

        if len(orientations) == len(repeating_unit):
            orientations = orientations*num_repeating_units

        chain_length = len(repeating_unit)*num_repeating_units
        if len(orientations) != chain_length:
            raise ValueError(
                'The length of orientations must match either '
                'the length of repeating_unit or the '
                'total number of vertices.'
            )

        generator = np.random.RandomState(random_seed)

        self._repeating_unit = self._normalize_repeating_unit(
            repeating_unit=repeating_unit
        )
        self._num_repeating_units = num_repeating_units

        vertex_data = [_AxleVertexData(0, 0, 0)]
        distance = 1 / (chain_length+1)
        choices = [True, False]
        for i, p in enumerate(orientations, 1):
            vertex_data.append(
                _CycleVertexData(
                    x=i*distance-0.5,
                    y=0,
                    z=0,
                    flip=generator.choice(choices, p=[p, 1-p])
                )
            )

        # Save the chosen orientations for __repr__.
        self._orientations = tuple(
            int(v.flip) for v in vertex_data[1:]
        )

        super().__init__(tuple(vertex_data), (), (), num_processes)

    @staticmethod
    def _normalize_repeating_unit(repeating_unit):
        if isinstance(repeating_unit, tuple):
            return repeating_unit
        base = ord('A')-1
        return tuple(ord(letter)-base for letter in repeating_unit)

    def assign_building_blocks_to_vertices(self, building_blocks):
        """
        Assign `building_blocks` to :attr:`vertices`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        Returns
        -------
        :class:`dict`
            Maps the `building_blocks`, to the
            :class:`~.topologies.base.Vertex` objects in
            :attr:`vertices` they are placed on during construction.
            The :class:`dict` has the form

            .. code-block:: python

                building_block_vertices = {
                    BuildingBlock(...): [Vertex(...), Vertex(...)],
                    BuildingBlock(...): [
                        Vertex(...),
                        Vertex(...),
                        Vertex(...),
                    ]
                    ConstructedMolecule(...): [Vertex(...)]
                }

        """

        threads = self._repeating_unit*self._num_repeating_units
        axle, *cycles = building_blocks
        building_block_vertices = {}
        building_block_vertices[axle] = self.vertices[0:1]
        for bb_index, vertex in zip(threads, self.vertices[1:]):
            bb = building_blocks[bb_index]
            building_block_vertices[bb] = (
                building_block_vertices.get(bb, [])
            )
            building_block_vertices[bb].append(vertex)
        return building_block_vertices

    def _get_scale(self, mol):
        """
        Get the scale used for the positions of :attr:`vertices`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        """

        axle = next(iter(mol.building_block_vertices))
        return 0.8*axle.get_maximum_diameter()

    def __repr__(self):
        return (
            f'rotaxane.NRotaxane('
            f'{self._repeating_unit!r}, '
            f'{self._num_repeating_units}, '
            f'{self._orientations!r}'
            f')'
        )



"""
Defines cages made from building blocks with 3 and 4 functional groups.

"""

from .base import Cage, _CageVertexData
from ..topology_graph import EdgeData


class SixPlusEight(Cage):
    """
    Represents a cage topology graph.

    Building blocks with three and four functional groups are required
    for this topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _x = 1
    _vertex_data = (
        _CageVertexData(-_x, _x, 0),
        _CageVertexData(-_x, -_x, 0),
        _CageVertexData(_x, _x, 0),
        _CageVertexData(_x, -_x, 0),

        _CageVertexData(0, 0, _x),
        _CageVertexData(0, 0, -_x),
    )

    vertex_data = (
        *_vertex_data,
        _CageVertexData.init_at_center(
            _vertex_data[0],
            _vertex_data[4],
            _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1],
            _vertex_data[4],
            _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4],
            _vertex_data[3],
            _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4],
            _vertex_data[2],
            _vertex_data[0]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[0],
            _vertex_data[5],
            _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5],
            _vertex_data[1],
            _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3],
            _vertex_data[5],
            _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2],
            _vertex_data[5],
            _vertex_data[0]
        )
    )

    edge_data = (
        EdgeData(vertex_data[6], vertex_data[0]),
        EdgeData(vertex_data[6], vertex_data[4]),
        EdgeData(vertex_data[6], vertex_data[1]),

        EdgeData(vertex_data[7], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[4]),
        EdgeData(vertex_data[7], vertex_data[3]),

        EdgeData(vertex_data[8], vertex_data[4]),
        EdgeData(vertex_data[8], vertex_data[3]),
        EdgeData(vertex_data[8], vertex_data[2]),

        EdgeData(vertex_data[9], vertex_data[4]),
        EdgeData(vertex_data[9], vertex_data[2]),
        EdgeData(vertex_data[9], vertex_data[0]),

        EdgeData(vertex_data[10], vertex_data[0]),
        EdgeData(vertex_data[10], vertex_data[5]),
        EdgeData(vertex_data[10], vertex_data[1]),

        EdgeData(vertex_data[11], vertex_data[5]),
        EdgeData(vertex_data[11], vertex_data[1]),
        EdgeData(vertex_data[11], vertex_data[3]),

        EdgeData(vertex_data[12], vertex_data[3]),
        EdgeData(vertex_data[12], vertex_data[5]),
        EdgeData(vertex_data[12], vertex_data[2]),

        EdgeData(vertex_data[13], vertex_data[2]),
        EdgeData(vertex_data[13], vertex_data[5]),
        EdgeData(vertex_data[13], vertex_data[0]),

    )

    num_windows = 12
    num_window_types = 1



"""
Defines electronic property calculators.

Electronic property calculators calculate the electronic properties of
molecules, namely the dipole moment, electron affinity and ionization
potential. Each :class:`.ElectronicPropertyCalculator` is initialized
with some settings and calculates the electronic properties of a
molecule with :meth:`~.ElectronicPropertyCalculator.dipole_moment`,
:meth:`~.ElectronicPropertyCalculator.electron_affinity` and
:meth:`~.ElectronicPropertyCalculator.ionization_potential`.

.. code-block:: python

    # Electronic property calculators work with any Molecule objects ,
    # such as BuildingBlock, Polymer, Cage etc.
    mol1 = BuildingBlock(...)
    mol2 = Cage(...)
    mol3 = Polymer(...)

    # Create the electronic property calculator.
    gfnxtb = GFNXTBElectronicProperties(gfnxtb_path='/opt/xtb/bin/xtb')
    mol1_ea = gfnxtb.electron_affinity(mol1)
    # We can optionally specify a conformer.
    mol2_ip = gfnxtb.ionization_potential(mol2, conformer=1)
    mol3_dm = gfnxtb.dipole_moment(mol3)

By default, calling any of the :class:`.ElectronicPropertyCalculator`
methods twice on the same molecule will calculate the property a
second time. However, we can use the
:attr:`~.ElectronicPropertyCalculator.use_cache` option to prevent
recalculations when the same molecule and conformer are given to the
same electronic property calculator a second time.

.. code-block:: python

    caching_gfnxtb = GFNXTBElectronicProperties(
        gfnxtb_path='/opt/xtb/bin/xtb',
        use_cache=True
    )
    # Calculate the EA the first time.
    ea1 = caching_gfnxtb.electron_affinity(mol1)
    # The second time, the EA is returned directly from memory, a
    # second calculation in not run.
    ea2 = caching_gfnxtb.electron_affinity(mol1)


Available electronic prperty calculators.
-----------------------------------------



.. _`adding electronic property calculators`:

Extending stk: Making new electronic property calculators.
----------------------------------------------------------

New electronic property calculators can be made by simply making a
class which inherits the :class:`.ElectronicPropertyCalculator` class.
In addition to this, the new class must define
:meth:`~.ElectronicPropertyCalculator.dopole_moment`,
:meth:`~.ElectronicPropertyCalculator.electron_affinity` and
:meth:`~.ElectronicPropertyCalculator.ionization_potential` methods.
Each method must take 2 arguments, a mandatory `mol` argument and an
optional `conformer` argument. The method will then calculate and
return the relevant property. There are no requirements regarding how
it should go about calculating the property. New energy calculators can
be added into the :mod:`.electronic_property_calculators` submodule or
into a new submodule.

"""

from .electronic_property_calculators import *
from .mopac import *



"""
Host Guest Complex
==================

#. :class:`.Complex`

"""


import numpy as np

from .topology_graph import TopologyGraph, VertexData, Vertex


class _HostVertexData(VertexData):
    def get_vertex(self):
        return _HostVertex(self)


class _HostVertex(Vertex):
    """
    Places the host in a :class:`.Complex`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def place_building_block(self, building_block, vertices, edges):
        building_block.set_centroid(self._position)
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        return dict()


class _GuestVertexData(VertexData):
    """
    Holds data for a :class:`_GuestVertex`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. Must match the index in
        :attr:`TopologyGraph.vertices`.

    position : :class:`numpy.ndarray`
        The position of the vertex.

    edges : :class:`list` of :class:`.EdgeData`
        The edges connected to the vertex.

    cell : :class:`numpy.ndarray`
        The unit cell in which the vertex is found.

    start : :class:`numpy.ndarray`
        A direction vector which gets aligned with `target`.

    target : :class:`numpy.ndarray`
        A direction vector which determines the rotation applied to
        the placed building block. A rotation such that
        `start` is transformed into `target` is applied
        to the placed building block.

    """

    def __init__(self, x, y, z, start, target):
        """
        Initialize a :class:`_GuestVertexData` instance.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

        start : :class:`numpy.ndarray`
            A direction vector which gets aligned with `target`.

        target : :class:`numpy.ndarray`
            A direction vector which determines the rotation applied to
            the placed building block. A rotation such that
            `start` is transformed into `target` is applied
            to the placed building block.

        """

        self.start = start
        self.target = target
        super().__init__(x, y, z)

    def clone(self, clear_edges=False):
        clone = super().clone(clear_edges)
        clone.start = np.array(self.start)
        clone.target = np.array(self.target)
        return clone

    def get_vertex(self):
        return _GuestVertex(self)


class _GuestVertex(Vertex):
    """
    Places the guest in a :class:`.Complex`.

    Attributes
    ----------
    id : :class:`int`
        The id of the vertex. This should be its index in
        :attr:`TopologyGraph.vertices`.

    """

    def __init__(self, data):
        self._start = np.array(data.start)
        self._target = np.array(data.target)
        super().__init__(data)

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            If ``True`` the :attr:`edges` attribute of the clone will
            be empty.

        Returns
        -------
        :class:`Vertex`
            The clone.

        """

        clone = super().clone(clear_edges)
        clone._start = np.array(self._start)
        clone._target = np.array(self._target)
        return clone

    def place_building_block(self, building_block, vertices, edges):
        building_block.set_centroid(self._position)
        building_block.apply_rotation_between_vectors(
            start=self._start,
            target=self._target,
            origin=self._position
        )
        return building_block.get_position_matrix()

    def assign_func_groups_to_edges(
        self,
        building_block,
        vertices,
        edges
    ):
        return dict()


class Complex(TopologyGraph):
    """
    Represents a host-guest complex topology graph.

    When using this topology graph, the host must be first in the
    `building_blocks` of the :class:`.ConstructedMolecule`
    and the guest must be second.

    Attributes
    ----------
    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    Examples
    --------
    .. code-block:: python

        import stk

        guest = stk.BuildingBlock('[Br][Br]')
        host = stk.ConstructedMolecule(
            building_blocks=[
                stk.BuildingBlock('NCCN', ['amine']),
                stk.BuildingBlock('O=CC(C=O)C=O', ['aldehyde'])
            ],
            topology_graph=stk.cage.FourPlusSix()
        )
        complex1 = stk.ConstructedMolecule(
            building_blocks=[host, guest],
            topology_graph=stk.host_guest.Complex()
        )

    Change the position and orientation of the guest

    .. code-block:: python

        complex2 = stk.ConstructedMolecule(
            building_blocks=[host, guest],
            topology_graph=stk.host_guest.Complex(
                # Apply a rotation onto the guest molecule such that
                # the vector returned by get_direction() has the same
                # direction as [1, 1, 1].
                guest_start=guest.get_direction(),
                guest_target=[1, 1, 1],
                displacement=[5.3, 2.1, 7.1]
            )
        )

    """

    def __init__(
        self,
        guest_start=None,
        guest_target=None,
        displacement=None,
        num_processes=1
    ):
        """
        Initialize an instance of :class:`.Complex`.

        Parameters
        ----------
        guest_start : :class:`tuple` of :class:`float`, optional
            A direction vector which gets aligned with `guest_target`.

        guest_target : :class:`tuple` of :class:`float`, optional
            A direction vector which determines the rotation applied to
            the guest building block. A rotation such that
            `guest_start` is transformed into `guest_target` is applied
            to the guest building block.

        displacement : :class:`list` of :class:`float`, optional
            The translational offset of the guest from the center of
            the host cavity.

        num_processes : :class:`int`, optional
            The number of parallel processes to create during
            :meth:`construct`.

        Raises
        ------
        :class:`TypeError`
            If `guest_start` or `guest_target` is defined but the other
            is not.

        """

        num_nones = sum(
            1 for vector in (guest_start, guest_target)
            if vector is None
        )
        if num_nones == 1:
            raise TypeError(
                'If guest_start or guest_target is defined, '
                'the other must be too.'
            )

        if guest_start is not None:
            start = guest_start = tuple(guest_start)
            target = guest_target = tuple(guest_target)
        else:
            start = target = (1., 0., 0.)

        # Save the values as None, for __repr__.
        self._guest_start = guest_start
        self._guest_target = guest_target
        self._displacement = displacement

        if displacement is None:
            displacement = np.array([0, 0, 0])

        x, y, z = displacement
        vertices = (
            _HostVertexData(0, 0, 0),
            _GuestVertexData(x, y, z, start, target)
        )
        super().__init__(vertices, (), (), num_processes)

    def assign_building_blocks_to_vertices(self, building_blocks):
        """
        Assign `building_blocks` to :attr:`vertices`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        Returns
        -------
        :class:`dict`
            Maps the `building_blocks`, to the
            :class:`~.topologies.base.Vertex` objects in
            :attr:`vertices` they are placed on during construction.
            The :class:`dict` has the form

            .. code-block:: python

                building_block_vertices = {
                    BuildingBlock(...): [Vertex(...), Vertex(...)],
                    BuildingBlock(...): [
                        Vertex(...),
                        Vertex(...),
                        Vertex(...),
                    ]
                    ConstructedMolecule(...): [Vertex(...)]
                }

        """

        return {
            bb: [vertex]
            for bb, vertex in zip(building_blocks, self.vertices)
        }

    def _get_scale(self, mol):
        """
        Get the scale used for the positions of :attr:`vertices`.

        Parameters
        ----------
        mol : :class:`.ConstructedMolecule`
            The molecule being constructed.

        Returns
        -------
        :class:`float` or :class:`list` of :class:`float`
            The value by which the position of each :class:`Vertex` is
            scaled. Can be a single number if all axes are scaled by
            the same amount or a :class:`list` of three numbers if
            each axis is scaled by a different value.

        """

        return 1

    def __repr__(self):
        return (
            f'host_guest.Complex('
            f'guest_start={self._guest_start!r}, '
            f'guest_target={self._guest_target!r}, '
            f'displacement={self._displacement!r})'
        )



"""
Defines cages made from building blocks with 2 and 4 functional groups.

"""


import numpy as np

from .base import Cage,  _CageVertexData
from ..topology_graph import EdgeData


class TwoPlusFour(Cage):
    """
    Represents a capsule cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    vertex_data = (
        _CageVertexData(0, 0, -1),
        _CageVertexData(0, 0, 1),

        _CageVertexData(2, 0, 0, False),
        _CageVertexData(-2, 0, 0, False),
        _CageVertexData(0, 2, 0, False),
        _CageVertexData(0, -2, 0, False)
    )

    edge_data = (
        EdgeData(vertex_data[2], vertex_data[0]),
        EdgeData(vertex_data[2], vertex_data[1]),
        EdgeData(vertex_data[3], vertex_data[0]),
        EdgeData(vertex_data[3], vertex_data[1]),
        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(vertex_data[4], vertex_data[1]),
        EdgeData(vertex_data[5], vertex_data[0]),
        EdgeData(vertex_data[5], vertex_data[1])
    )

    num_windows = 4
    num_window_types = 1


class ThreePlusSix(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _x = 1
    vertex_data = (
        _CageVertexData(-2*_x, -_x*np.sqrt(3), 0, False),
        _CageVertexData(2*_x, -_x*np.sqrt(3), 0, False),
        _CageVertexData(0, _x*np.sqrt(3), 0, False),

        _CageVertexData(0, -2*_x*np.sqrt(3), _x, False),
        _CageVertexData(0, -2*_x*np.sqrt(3), -_x, False),

        _CageVertexData(2*_x, 0, _x, False),
        _CageVertexData(2*_x, 0, -_x, False),

        _CageVertexData(-2*_x, 0, _x, False),
        _CageVertexData(-2*_x, 0, -_x, False),
    )

    edge_data = (
        EdgeData(vertex_data[3], vertex_data[0]),
        EdgeData(vertex_data[3], vertex_data[1]),

        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(vertex_data[4], vertex_data[1]),

        EdgeData(vertex_data[5], vertex_data[1]),
        EdgeData(vertex_data[5], vertex_data[2]),

        EdgeData(vertex_data[6], vertex_data[1]),
        EdgeData(vertex_data[6], vertex_data[2]),

        EdgeData(vertex_data[7], vertex_data[0]),
        EdgeData(vertex_data[7], vertex_data[2]),

        EdgeData(vertex_data[8], vertex_data[0]),
        EdgeData(vertex_data[8], vertex_data[2]),

    )

    num_windows = 5
    num_window_types = 2


class FourPlusEight(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    vertex_data = (
        _CageVertexData(-1, -1, 0, False),
        _CageVertexData(-1, 1, 0, False),

        _CageVertexData(1, -1, 0, False),
        _CageVertexData(1, 1, 0, False),

        _CageVertexData(-2, 0, 1, False),
        _CageVertexData(-2, 0, -1, False),

        _CageVertexData(0, 2, 1, False),
        _CageVertexData(0, 2, -1, False),

        _CageVertexData(0, -2, 1, False),
        _CageVertexData(0, -2, -1, False),

        _CageVertexData(2, 0, 1, False),
        _CageVertexData(2, 0, -1, False)

    )

    edge_data = (
        EdgeData(vertex_data[4], vertex_data[0]),
        EdgeData(vertex_data[4], vertex_data[1]),

        EdgeData(vertex_data[5], vertex_data[0]),
        EdgeData(vertex_data[5], vertex_data[1]),

        EdgeData(vertex_data[6], vertex_data[1]),
        EdgeData(vertex_data[6], vertex_data[3]),

        EdgeData(vertex_data[7], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[3]),

        EdgeData(vertex_data[8], vertex_data[0]),
        EdgeData(vertex_data[8], vertex_data[2]),

        EdgeData(vertex_data[9], vertex_data[0]),
        EdgeData(vertex_data[9], vertex_data[2]),

        EdgeData(vertex_data[10], vertex_data[2]),
        EdgeData(vertex_data[10], vertex_data[3]),

        EdgeData(vertex_data[11], vertex_data[2]),
        EdgeData(vertex_data[11], vertex_data[3]),

    )

    num_windows = 6
    num_window_types = 2


class FivePlusTen(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.
    """

    _c1 = np.cos(2*np.pi/5)
    _c2 = np.cos(np.pi/5)
    _s1 = np.sin(2*np.pi/5)
    _s2 = np.sin(4*np.pi/5)

    vertex_data = (
        _CageVertexData(0, 1, 0, False),
        _CageVertexData(_s1, _c1, 0, False),
        _CageVertexData(_s2, -_c2, 0, False),

        _CageVertexData(-_s2, -_c2, 0, False),
        _CageVertexData(-_s1, _c1, 0, False),

        _CageVertexData(_s1, 1+_c1, 0.5, False),
        _CageVertexData(_s1, 1+_c1, -0.5, False),

        _CageVertexData(_s1+_s2, _c1-_c2, 0.5, False),
        _CageVertexData(_s1+_s2, _c1-_c2, -0.5, False),

        _CageVertexData(0, -2*_c2, 0.5, False),
        _CageVertexData(0, -2*_c2, -0.5, False),

        _CageVertexData(-_s2-_s1, -_c2+_c1, 0.5, False),
        _CageVertexData(-_s2-_s1, -_c2+_c1, -0.5, False),

        _CageVertexData(-_s1, 1+_c1, 0.5, False),
        _CageVertexData(-_s1, 1+_c1, -0.5, False),

    )

    edge_data = (
        EdgeData(vertex_data[5], vertex_data[0]),
        EdgeData(vertex_data[5], vertex_data[1]),
        EdgeData(vertex_data[6], vertex_data[0]),
        EdgeData(vertex_data[6], vertex_data[1]),

        EdgeData(vertex_data[7], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[2]),
        EdgeData(vertex_data[8], vertex_data[1]),
        EdgeData(vertex_data[8], vertex_data[2]),

        EdgeData(vertex_data[9], vertex_data[2]),
        EdgeData(vertex_data[9], vertex_data[3]),
        EdgeData(vertex_data[10], vertex_data[2]),
        EdgeData(vertex_data[10], vertex_data[3]),

        EdgeData(vertex_data[11], vertex_data[3]),
        EdgeData(vertex_data[11], vertex_data[4]),
        EdgeData(vertex_data[12], vertex_data[3]),
        EdgeData(vertex_data[12], vertex_data[4]),

        EdgeData(vertex_data[13], vertex_data[4]),
        EdgeData(vertex_data[13], vertex_data[0]),
        EdgeData(vertex_data[14], vertex_data[4]),
        EdgeData(vertex_data[14], vertex_data[0])

    )

    num_windows = 7
    num_window_types = 2


class SixPlusTwelve(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _vertex_data = (
        _CageVertexData(-1, -1, 0),
        _CageVertexData(-1, 1, 0),
        _CageVertexData(1, -1, 0),
        _CageVertexData(1, 1, 0),
        _CageVertexData(0, 0, 1),
        _CageVertexData(0, 0, -1)
    )

    vertex_data = (
        *_vertex_data,
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[2]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[0]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[1]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[0]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[3]
        )
    )

    edge_data = (
        EdgeData(vertex_data[6], vertex_data[0]),
        EdgeData(vertex_data[6], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[1]),
        EdgeData(vertex_data[7], vertex_data[3]),
        EdgeData(vertex_data[8], vertex_data[3]),
        EdgeData(vertex_data[8], vertex_data[2]),

        EdgeData(vertex_data[9], vertex_data[0]),
        EdgeData(vertex_data[9], vertex_data[2]),
        EdgeData(vertex_data[10], vertex_data[4]),
        EdgeData(vertex_data[10], vertex_data[0]),
        EdgeData(vertex_data[11], vertex_data[4]),
        EdgeData(vertex_data[11], vertex_data[1]),

        EdgeData(vertex_data[12], vertex_data[4]),
        EdgeData(vertex_data[12], vertex_data[2]),
        EdgeData(vertex_data[13], vertex_data[4]),
        EdgeData(vertex_data[13], vertex_data[3]),
        EdgeData(vertex_data[14], vertex_data[5]),
        EdgeData(vertex_data[14], vertex_data[0]),

        EdgeData(vertex_data[15], vertex_data[5]),
        EdgeData(vertex_data[15], vertex_data[1]),
        EdgeData(vertex_data[16], vertex_data[5]),
        EdgeData(vertex_data[16], vertex_data[2]),
        EdgeData(vertex_data[17], vertex_data[5]),
        EdgeData(vertex_data[17], vertex_data[3])
    )

    num_windows = 8
    num_window_types = 1


class EightPlusSixteen(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _x = 2
    _vertex_data = (
        _CageVertexData(-0.5*_x, 0.5*_x, -0.35*_x),
        _CageVertexData(-0.5*_x, -0.5*_x, -0.35*_x),
        _CageVertexData(0.5*_x, -0.5*_x, -0.35*_x),
        _CageVertexData(0.5*_x, 0.5*_x, -0.35*_x),

        _CageVertexData(-_x*np.sqrt(2)/2, 0, _x*0.35),
        _CageVertexData(0, -_x*np.sqrt(2)/2, _x*0.35),
        _CageVertexData(_x*np.sqrt(2)/2, 0, _x*0.35),
        _CageVertexData(0, _x*np.sqrt(2)/2, _x*0.35)
    )

    vertex_data = (
        *_vertex_data,

        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[4]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[7]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[0]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[6], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[7], _vertex_data[4]
        ),

    )

    edge_data = (
        EdgeData(vertex_data[8], vertex_data[1]),
        EdgeData(vertex_data[8], vertex_data[5]),

        EdgeData(vertex_data[9], vertex_data[2]),
        EdgeData(vertex_data[9], vertex_data[5]),

        EdgeData(vertex_data[10], vertex_data[0]),
        EdgeData(vertex_data[10], vertex_data[4]),

        EdgeData(vertex_data[11], vertex_data[1]),
        EdgeData(vertex_data[11], vertex_data[4]),

        EdgeData(vertex_data[12], vertex_data[2]),
        EdgeData(vertex_data[12], vertex_data[6]),

        EdgeData(vertex_data[13], vertex_data[3]),
        EdgeData(vertex_data[13], vertex_data[6]),

        EdgeData(vertex_data[14], vertex_data[0]),
        EdgeData(vertex_data[14], vertex_data[7]),

        EdgeData(vertex_data[15], vertex_data[3]),
        EdgeData(vertex_data[15], vertex_data[7]),

        EdgeData(vertex_data[16], vertex_data[0]),
        EdgeData(vertex_data[16], vertex_data[1]),

        EdgeData(vertex_data[17], vertex_data[1]),
        EdgeData(vertex_data[17], vertex_data[2]),

        EdgeData(vertex_data[18], vertex_data[2]),
        EdgeData(vertex_data[18], vertex_data[3]),

        EdgeData(vertex_data[19], vertex_data[3]),
        EdgeData(vertex_data[19], vertex_data[0]),

        EdgeData(vertex_data[20], vertex_data[4]),
        EdgeData(vertex_data[20], vertex_data[5]),

        EdgeData(vertex_data[21], vertex_data[5]),
        EdgeData(vertex_data[21], vertex_data[6]),

        EdgeData(vertex_data[22], vertex_data[6]),
        EdgeData(vertex_data[22], vertex_data[7]),

        EdgeData(vertex_data[23], vertex_data[7]),
        EdgeData(vertex_data[23], vertex_data[4])
    )

    num_windows = 10
    num_window_types = 2


class TenPlusTwenty(Cage):
    """
    Represents a cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _x = 2
    _vertex_data = (
        _CageVertexData(-_x, _x, -_x),
        _CageVertexData(-_x, -_x, -_x),
        _CageVertexData(_x, _x, -_x),
        _CageVertexData(_x, -_x, -_x),

        _CageVertexData(-_x, _x, _x),
        _CageVertexData(-_x, -_x, _x),
        _CageVertexData(_x, _x, _x),
        _CageVertexData(_x, -_x, _x),

        _CageVertexData(0, 0, _x*1.5),
        _CageVertexData(0, 0, -_x*1.5)
    )

    vertex_data = (
        *_vertex_data,

        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[3]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[3]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[4], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[5], _vertex_data[7]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[6], _vertex_data[7]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[0], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[1], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[2], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[3], _vertex_data[7]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[4]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[5]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[6]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[8], _vertex_data[7]
        ),

        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[0]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[1]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[2]
        ),
        _CageVertexData.init_at_center(
            _vertex_data[9], _vertex_data[3]
        )
    )

    edge_data = (
        EdgeData(vertex_data[10], vertex_data[0]),
        EdgeData(vertex_data[10], vertex_data[2]),
        EdgeData(vertex_data[11], vertex_data[0]),
        EdgeData(vertex_data[11], vertex_data[1]),

        EdgeData(vertex_data[12], vertex_data[1]),
        EdgeData(vertex_data[12], vertex_data[3]),
        EdgeData(vertex_data[13], vertex_data[2]),
        EdgeData(vertex_data[13], vertex_data[3]),

        EdgeData(vertex_data[14], vertex_data[4]),
        EdgeData(vertex_data[14], vertex_data[6]),
        EdgeData(vertex_data[15], vertex_data[4]),
        EdgeData(vertex_data[15], vertex_data[5]),

        EdgeData(vertex_data[16], vertex_data[5]),
        EdgeData(vertex_data[16], vertex_data[7]),
        EdgeData(vertex_data[17], vertex_data[6]),
        EdgeData(vertex_data[17], vertex_data[7]),

        EdgeData(vertex_data[18], vertex_data[0]),
        EdgeData(vertex_data[18], vertex_data[4]),
        EdgeData(vertex_data[19], vertex_data[1]),
        EdgeData(vertex_data[19], vertex_data[5]),

        EdgeData(vertex_data[20], vertex_data[2]),
        EdgeData(vertex_data[20], vertex_data[6]),
        EdgeData(vertex_data[21], vertex_data[3]),
        EdgeData(vertex_data[21], vertex_data[7]),

        EdgeData(vertex_data[22], vertex_data[8]),
        EdgeData(vertex_data[22], vertex_data[4]),
        EdgeData(vertex_data[23], vertex_data[8]),
        EdgeData(vertex_data[23], vertex_data[5]),

        EdgeData(vertex_data[24], vertex_data[8]),
        EdgeData(vertex_data[24], vertex_data[6]),
        EdgeData(vertex_data[25], vertex_data[8]),
        EdgeData(vertex_data[25], vertex_data[7]),

        EdgeData(vertex_data[26], vertex_data[9]),
        EdgeData(vertex_data[26], vertex_data[0]),
        EdgeData(vertex_data[27], vertex_data[9]),
        EdgeData(vertex_data[27], vertex_data[1]),

        EdgeData(vertex_data[28], vertex_data[9]),
        EdgeData(vertex_data[28], vertex_data[2]),
        EdgeData(vertex_data[29], vertex_data[9]),
        EdgeData(vertex_data[29], vertex_data[3])
    )

    num_windows = 12
    num_window_types = 2



"""
Defines cages made from building blocks with 3 functional groups.

"""

import numpy as np

from .base import Cage,  _CageVertexData, _CageVertex
from ..topology_graph import EdgeData


class _OnePlusOneVertexData(_CageVertexData):
    def __init__(
        self,
        x,
        y,
        z,
        edge_normal,
        use_bonder_placement=True
    ):
        """
        Initialize a :class:`_OnePlusOneVertexData` instance.

        Parameters
        ----------
        x : :class:`float`
            The x coordinate.

        y : :class:`float`
            The y coordinate.

        z : :class:`float`
            The z coordinate.

        edge_normal : :class:`list` of :class:`int`
            The edge plane normal to use.

        use_bonder_placement : :class:`bool`, optional
            If ``True``the position of the vertex will be updated such
            that it is in the middle of the neighboring bonder
            centroids, rather than in the middle of the neighboring
            vertices.

        """

        self.edge_normal = edge_normal
        super().__init__(
            x=x,
            y=y,
            z=z,
            use_bonder_placement=use_bonder_placement
        )

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            ``True`` if the clone should not be connected to any edges.

        Returns
        -------
        :class:`_OnePlusOneVertexData`
            The clone.

        """

        clone = super().clone(clear_edges)
        clone.edge_normal = list(self.edge_normal)
        return clone

    def get_vertex(self):
        return _OnePlusOneVertex(self)


class _OnePlusOneVertex(_CageVertex):

    def __init__(self, data):
        self._edge_normal = list(data.edge_normal)
        super().__init__(data)

    def clone(self, clear_edges=False):
        """
        Return a clone.

        Parameters
        ----------
        clear_edges : :class:`bool`, optional
            ``True`` if the clone should not be connected to any edges.

        Returns
        -------
        :class:`Vertex`
            The clone.

        """

        clone = super().clone(clear_edges)
        clone._edge_normal = list(self._edge_normal)
        return clone

    def _place_nonlinear_building_block(
        self,
        building_block,
        vertices,
        edges
    ):
        """
        Place `building_block` on the :class:`.Vertex`.

        Parameters
        ----------
        building_block : :class:`.BuildingBlock`
            The building block molecule which is to be placed on the
            vertex.

        vertices : :class:`tuple` of :class:`.Vertex`
            All vertices in the topology graph. The index of each
            vertex must match its :class:`~.Vertex.id`.

        edges : :class:`tuple` of :class:`.Edge`
            All edges in the topology graph. The index of each
            edge must match its :class:`~.Edge.id`.

        Returns
        -------
        :class:`numpy.nadarray`
            The position matrix of `building_block` after being
            placed.

        """

        building_block.set_centroid(
            position=self._position,
            atom_ids=building_block.get_bonder_ids()
        )
        building_block.apply_rotation_between_vectors(
            start=building_block.get_bonder_plane_normal(),
            target=self._edge_normal,
            origin=self._position
        )
        fg_bonder_centroid = building_block.get_centroid(
            atom_ids=building_block.func_groups[0].get_bonder_ids()
        )
        start = fg_bonder_centroid - self._position
        aligner_edge = edges[self._edge_ids[self._aligner_edge]]
        edge_coord = aligner_edge.get_position()
        connected_edges = tuple(edges[id_] for id_ in self._edge_ids)
        target = edge_coord - self._get_edge_centroid(
            centroid_edges=connected_edges,
            vertices=vertices
        )
        building_block.apply_rotation_to_minimize_angle(
            start=start,
            target=target,
            axis=self._edge_normal,
            origin=self._position
        )
        return building_block.get_position_matrix()


class OnePlusOne(Cage):
    """
    Represents a capsule cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _x = 1
    vertex_data = (
        _OnePlusOneVertexData(_x, 0., 0., [1, 0, 0], False),
        _OnePlusOneVertexData(-_x, 0., 0., [-1, 0, 0], False),

    )
    edge_data = (
        EdgeData(
            vertex_data[0],
            vertex_data[1],
            position=np.array([0., 1., 0.])
        ),
        EdgeData(
            vertex_data[0],
            vertex_data[1],
            position=np.array([0., -1., 1.])
        ),
        EdgeData(
            vertex_data[0],
            vertex_data[1],
            position=np.array([0., -1., -1.])
        )
    )

    num_windows = 3
    num_window_types = 1


class TwoPlusTwo(Cage):
    """
    Represents a tetrahedron cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _x = 1
    vertex_data = (
        _CageVertexData(_x, 0, -_x/np.sqrt(2), False),
        _CageVertexData(-_x, 0, -_x/np.sqrt(2), False),
        _CageVertexData(0, _x, _x/np.sqrt(2), False),
        _CageVertexData(0, -_x, _x/np.sqrt(2), False)
    )

    edge_data = (
        EdgeData(vertex_data[0], vertex_data[1]),
        EdgeData(vertex_data[0], vertex_data[2]),
        EdgeData(vertex_data[0], vertex_data[3]),

        EdgeData(vertex_data[1], vertex_data[2]),
        EdgeData(vertex_data[1], vertex_data[3]),

        EdgeData(vertex_data[2], vertex_data[3])
    )

    num_windows = 4
    num_window_types = 1


class FourPlusFour(Cage):
    """
    Represents a cube cage topology graph.

    See :class:`.Cage` for more details and examples.

    Attributes
    ----------
    vertex_data : :class:`tuple` of :class:`.VertexData`
        A class attribute. Holds the data of the vertices which make up
        the topology graph.

    edge_data : :class:`tuple` of :class:`.EdgeData`
        A class attribute. Holds the data of the edges which make up
        the topology graph.

    vertices : :class:`tuple` of :class:`.Vertex`
        The vertices which make up the topology graph.

    edges : :class:`tuple` of :class:`.Edge`
        The edges which make up the topology graph.

    """

    _x = 1
    vertex_data = (
        _CageVertexData(-_x, _x, -_x, False),
        _CageVertexData(-_x, -_x, -_x, False),
        _CageVertexData(_x, _x, -_x, False),
        _CageVertexData(_x, -_x, -_x, False),

        _CageVertexData(-_x, _x, _x, False),
        _CageVertexData(-_x, -_x, _x, False),
        _CageVertexData(_x, _x, _x, False),
        _CageVertexData(_x, -_x, _x, False)
    )

    edge_data = (
        EdgeData(vertex_data[0], vertex_data[1]),
        EdgeData(vertex_data[0], vertex_data[2]),
        EdgeData(vertex_data[0], vertex_data[4]),
        EdgeData(vertex_data[1], vertex_data[3]),
        EdgeData(vertex_data[1], vertex_data[5]),
        EdgeData(vertex_data[2], vertex_data[6]),
        EdgeData(vertex_data[2], vertex_data[3]),
        EdgeData(vertex_data[3], vertex_data[7]),
        EdgeData(vertex_data[4], vertex_data[6]),
        EdgeData(vertex_data[4], vertex_data[5]),
        EdgeData(vertex_data[5], vertex_data[7]),
        EdgeData(vertex_data[6], vertex_data[7])
    )

    num_windows = 6
    num_window_types = 1



import json
import os
import numpy as np
import rdkit.Chem.AllChem as rdkit
from scipy.spatial.distance import euclidean

from ...utilities import (
    vector_angle,
    rotation_matrix,
    rotation_matrix_arbitrary_axis,
    mol_from_mae_file,
    remake,
    periodic_table
)


class MoleculeSubclassError(Exception):
    ...


class _Cached(type):
    def __call__(cls, *args, **kwargs):
        return cls._construct(*args, **kwargs)


class Molecule(metaclass=_Cached):
    """
    Represents molecules.

    This is an abstract base class defines the operations which any
    class describing molecules should inherit. Examples are
    :class:`.BuildingBlock` and :class:`.ConstructedMolecule`.
    This class should not be initialized directly, use an instance of a
    subclass instead.

    Attributes
    ----------
    atoms : :class:`tuple` of :class:`.Atom`
        The atoms of the molecule.

    bonds : :class:`tuple` of :class:`.Bond`
        The bonds of the molecule.

    """

    # Maps the name of each subclass to the class. Necessary for
    # loading molecules from dict represeentations, as the dict
    # representation will hold the name of the subclass.
    _subclasses = {}

    @classmethod
    def _construct(cls, *args, **kwargs):
        """
        Construct a new instance.

        When a :class:`.Molecule` or its subclass is created with the
        default initialzer, this method is called instead of
        :meth:`__init__`.

        .. code-block:: python

            import stk

            # bb is the return value of BuildingBlock._construct()
            bb = stk.BuildingBlock('NCCN')

        Because of this, :meth:`_construct` takes the same paramters
        as :meth:`__init__`.

        This method is necessary so that the default initializer can
        use caching efficiently. For example, in order to get the
        identity key of a :class:`.BuildingBlock`, the SMILES received
        need to be converted into an :mod:`rdkit` molecule, which is
        then used to generate canonical SMILES, which are part of the
        identity key. If the :class:`.BuildingBlock` is not cached,
        the SMILES would then need to be passed to
        :meth:`.BuildingBlock.__init__` so that the new
        :class:`.BuildingBlock` instance can be made. However,
        :meth:`.BuildingBlock.__init__` creates an :mod:`rdkit`
        molecule all over again, which is useless duplication.

        By using :meth:`._construct` the implementation can avoid
        calling :meth:`.BuildingBlock.__init__` if an :mod:`rdkit`
        molecule has already been made. Instead,
        :meth:`.BuildingBlock.init_from_rdkit_mol` can be used.

        Parameters
        ----------
        *args : :class:`object`
            The arguments passed to :methd`__init__`.

        **kwargs : :class:`object`
            The keyword arguments passed to :meth:`__init__`.

        Returns
        -------
        :class:`.Molecule`
            The new instance.

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method which must be implemented in a
            subclass.

        """

        raise NotImplementedError()

    def __init__(self, atoms, bonds, position_matrix, identity_key):
        """
        Initialize a :class:`Molecule`.

        Parameters
        ----------
        atoms : :class:`tuple` of :class:`.Atom`
            The atoms which compose the molecule.

        bonds : :class:`tuple` of :class:`.Bond`
            The bonds of the molecule.

        position_matrix : :class:`numpy.ndarray`
            A ``(n, 3)`` matrix holding the position of every atom in
            the :class:`.Molecule`.

        identity_key : :class:`object`
            The identity key of the molecule. Molecules which
            ``stk`` sees as identical will have the same identity key.
            Must be hashable.

        """

        self.atoms = atoms
        self.bonds = bonds
        # A (3, n) numpy.ndarray holding the position of every atom in
        # the molecule.
        self._position_matrix = position_matrix.T
        self._identity_key = identity_key

    @classmethod
    def init_from_dict(self, mol_dict, use_cache=False):
        """
        Initialize from a :class:`dict` representation.

        The :class:`Molecule` returned has the class specified in
        `mol_dict`, not :class:`Molecule`.

        Parameters
        ----------
        mol_dict : :class:`dict`
            A :class:`dict` holding the :class:`dict` representation
            of a molecule, generated by :meth:`to_dict`.

        use_cache : :class:`bool`, optional
            If ``True``, a new instance will not be made if a cached
            and identical one already exists, the one which already
            exists will be returned. If ``True`` and a cached,
            identical instance does not yet exist the created one will
            be added to the cache.

        Returns
        -------
        :class:`Molecule`
            The molecule represented by `mol_dict`.

        """

        c = self._subclasses[mol_dict['class']]
        return c._init_from_dict(mol_dict, use_cache=use_cache)

    @classmethod
    def _init_from_dict(cls, mol_dict, use_cache):
        """
        Initialize from a :class:`dict` representation.

        Parameters
        ----------
        mol_dict : :class:`dict`
            A :class:`dict` representation of a molecule generated
            by :meth:`to_dict`.

        use_cache : :class:`bool`, optional
            If ``True``, a new instance will not be made if a cached
            and identical one already exists, the one which already
            exists will be returned. If ``True`` and a cached,
            identical instance does not yet exist the created one will
            be added to the cache.

        Returns
        -------
        :class:`dict`
            A :class:`dict` representation.

        Raises
        ------
        :class:`NotImplementedError`
            This method needs to be implemented by a subclass.

        """

        raise NotImplementedError()

    def __init_subclass__(cls, **kwargs):
        if cls.__name__ in cls._subclasses:
            msg = 'Subclass with this name already exists.'
            raise MoleculeSubclassError(msg)
        cls._subclasses[cls.__name__] = cls
        # Maps the _key of each instance of a subclass to the
        # to the actual instance.
        cls._cache = {}
        super().__init_subclass__(**kwargs)

    def apply_displacement(self, displacement):
        """
        Shift the centroid by `displacement`.

        Parameters
        ----------
        displacement : :class:`numpy.ndarray`
            A displacement vector applied to the molecule.

        Returns
        -------
        :class:`.Molecule`
            The molecule.

        """

        self._position_matrix = (
            self._position_matrix.T + displacement
        ).T
        return self

    def apply_rotation_about_axis(self, angle, axis, origin):
        """
        Rotate by `angle` about `axis` on the `origin`.

        Parameters
        ----------
        angle : :class:`float`
            The size of the rotation in radians.

        axis : :class:`numpy.ndarray`
            The axis about which the rotation happens.

        origin : :class:`numpy.ndarray`
            The origin about which the rotation happens.

        Returns
        -------
        :class:`.Molecule`
            The molecule.

        """

        # Set the origin of the rotation to "origin".
        self.apply_displacement(-origin)
        rot_mat = rotation_matrix_arbitrary_axis(angle, axis)

        # Apply the rotation matrix on the position matrix, to get the
        # new position matrix.
        self._position_matrix = rot_mat @ self._position_matrix

        # Return the centroid of the molecule to the original position.
        self.apply_displacement(origin)
        return self

    def apply_rotation_between_vectors(self, start, target, origin):
        """
        Rotate by a rotation from `start` to `target`.

        Given two direction vectors, `start` and `target`, this method
        applies the rotation required transform `start` to `target`
        onto the molecule. The rotation occurs about the `origin`.

        For example, if the `start` and `target` vectors
        are 45 degrees apart, a 45 degree rotation will be applied to
        the molecule. The rotation will be along the appropriate
        direction.

        The great thing about this method is that you as long as you
        can associate a geometric feature of the molecule with a
        vector, then the molecule can be rotated so that this vector is
        aligned with `target`. The defined vector can be virtually
        anything. This means that any geometric feature of the molecule
        can be easily aligned with any arbitrary axis.

        Parameters
        ----------
        start : :class:`numpy.ndarray`
            A vector which is to be rotated so that it transforms into
            the `target` vector.

        target : :class:`numpy.ndarray`
            The vector onto which `start` is rotated.

        origin : :class:`numpy.ndarray`
            The point about which the rotation occurs.

        Returns
        -------
        :class:`.Molecule`
            The molecule.

        """

        # Set the origin of the rotation to "origin".
        self.apply_displacement(-origin)
        rot_mat = rotation_matrix(start, target)

        # Apply the rotation matrix to the atomic positions to yield
        # the new atomic positions.
        self._position_matrix = rot_mat @ self._position_matrix

        # Restore original position.
        self.apply_displacement(origin)
        return self

    def apply_rotation_to_minimize_angle(
        self,
        start,
        target,
        axis,
        origin
    ):
        """
        Rotate to minimize the angle between `start` and `target`.

        Note that this function will not necessarily overlay the
        `start` and `target` vectors. This is because the possible
        rotation is restricted to the `axis`.

        Parameters
        ----------
        start : :class:`numpy.ndarray`
            The vector which is rotated.

        target : :class:`numpy.ndarray`
            The vector which is stationary.

        axis : :class:`numpy.ndarray`
            The vector about which the rotation happens.

        origin : :class:`numpy.ndarray`
            The origin about which the rotation happens.

        Returns
        -------
        :class:`.Molecule`
            The molecule.

        """

        # If the vector being rotated is not finite, exit. This is
        # probably due to a planar molecule.
        if not all(np.isfinite(x) for x in start):
            return self

        self.apply_displacement(-origin)

        # 1. First transform the problem.
        # 2. The rotation axis is set equal to the z-axis.
        # 3. Apply this transformation to all vectors in the problem.
        # 4. Take only the x and y components of `start` and `target`.
        # 5. Work out the angle between them.
        # 6. Apply that rotation along the original rotation axis.

        rotmat = rotation_matrix(axis, [0, 0, 1])
        tstart = np.dot(rotmat, start)
        tstart = np.array([tstart[0], tstart[1], 0])

        # If the `tstart` vector is 0 after these transformations it
        # means that it is parallel to the rotation axis, stop.
        if np.allclose(tstart, [0, 0, 0], 1e-8):
            self.apply_displacement(origin)
            return self

        tend = np.dot(rotmat, target)
        tend = np.array([tend[0], tend[1], 0])
        angle = vector_angle(tstart, tend)

        # Check in which direction the rotation should go.
        # This is done by applying the rotation in each direction and
        # seeing which one leads to a smaller angle.
        r1 = rotation_matrix_arbitrary_axis(angle, [0, 0, 1])
        t1 = vector_angle(np.dot(r1, tstart), tend)
        r2 = rotation_matrix_arbitrary_axis(-angle, [0, 0, 1])
        t2 = vector_angle(np.dot(r2, tstart), tend)

        if t2 < t1:
            angle *= -1

        rot_mat = rotation_matrix_arbitrary_axis(angle, axis)
        self._position_matrix = rot_mat @ self._position_matrix
        self.apply_displacement(origin)
        return self

    def get_atom_positions(self, atom_ids=None):
        """
        Yield the positions of atoms.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of the atoms whose positions are desired.
            If ``None``, then the positions of all atoms will be
            yielded.

        Yields
        ------
        :class:`numpy.ndarray`
            The x, y and z coordinates of an atom.

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        coords = self._position_matrix[:, atom_ids].T
        for atom_coords in coords:
            yield atom_coords

    @classmethod
    def get_cached_mol(cls, identity_key, default=None):
        """
        Get a molecule from the cache.

        Parameters
        ----------
        identity_key : :class:`object`
            The identity key of the molecule to return.

        default : :class:`object`, optional
            Returned if `identity_key` is not found in the cache.
            If ``None`` an error will be raised if `identity_key` is
            not found in the cache.

        Returns
        -------
        :class:`.Molecule`
            The cached molecule.

        """

        if default is None:
            return cls._cache[identity_key]
        return cls._cache.get(identity_key, default)

    def get_atom_distance(self, atom1_id, atom2_id):
        """
        Return the distance between 2 atoms.

        This method does not account for the van der Waals radius of
        atoms.

        Parameters
        ----------
        atom1_id : :class:`int`
            The id of the first atom.

        atom2_id : :class:`int`
            The id of the second atom.

        Returns
        -------
        :class:`float`
            The distance between the first and second atoms.

        """

        distance = euclidean(
            u=self._position_matrix[:, atom1_id],
            v=self._position_matrix[:, atom2_id]
        )
        return float(distance)

    def get_center_of_mass(self, atom_ids=None):
        """
        Return the centre of mass.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of atoms which should be used to calculate the
            center of mass. If ``None``, then all atoms will be used.

        Returns
        -------
        :class:`numpy.ndarray`
            The coordinates of the center of mass.

        References
        ----------
        https://en.wikipedia.org/wiki/Center_of_mass

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))
        elif not isinstance(atom_ids, (list, tuple)):
            # Iterable gets used twice, once in get_atom_positions
            # and once in zip.
            atom_ids = list(atom_ids)

        center = 0
        total_mass = 0.
        coords = self.get_atom_positions(atom_ids)
        for atom_id, coord in zip(atom_ids, coords):
            mass = self.atoms[atom_id].mass
            total_mass += mass
            center += mass*coord
        return np.divide(center, total_mass)

    def get_centroid(self, atom_ids=None):
        """
        Return the centroid.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of atoms which are used to calculate the
            centroid. If ``None``, then all atoms will be used.

        Returns
        -------
        :class:`numpy.ndarray`
            The centroid of atoms specified by `atom_ids`.

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        return np.divide(
            self._position_matrix[:, atom_ids].sum(axis=1),
            len(atom_ids)
        )

    def get_direction(self, atom_ids=None):
        """
        Return a vector of best fit through the atoms.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of atoms which should be used to calculate the
            vector. If ``None``, then all atoms will be used.

        Returns
        -------
        :class:`numpy.ndarray`
            The vector of best fit.

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        pos = self._position_matrix[:, atom_ids].T
        return np.linalg.svd(pos - pos.mean(axis=0))[-1][0]

    def get_identity_key(self):
        """
        Return the identity key.

        The identity key wil be equal for two molecules which
        ``stk`` sees as identical. The identity key does not take
        the conformation into account but it does account for
        isomerism.

        Returns
        -------
        :class:`object`
            A hashable object which represents the identity of the
            molecule.

        """

        return self._identity_key

    def get_maximum_diameter(self, atom_ids=None):
        """
        Return the maximum diamater.

        This method does not account for the van der Waals radius of
        atoms.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`
            The ids of atoms which are considered when looking for the
            maximum diamater. If ``None`` then all atoms in the
            molecule are considered.

        Returns
        -------
        :class:`float`
            The maximum diameter in the molecule.

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))
        elif not isinstance(atom_ids, (list, tuple)):
            atom_ids = list(atom_ids)

        coords = self._position_matrix[:, atom_ids]
        return float(euclidean(coords.min(axis=1), coords.max(axis=1)))

    def get_plane_normal(self, atom_ids=None):
        """
        Return the normal to the plane of best fit.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The ids of atoms which should be used to calculate the
            plane. If ``None``, then all atoms will be used.

        Returns
        -------
        :class:`numpy.ndarray`
            Vector orthonormal to the plane of the molecule.

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))
        # atom_ids needs to be sorted so that the same reference vector
        # is always created for equivalent inputs.
        elif not isinstance(atom_ids, range):
            atom_ids = sorted(atom_ids)

        pos = self._position_matrix[:, atom_ids].T
        centroid = self.get_centroid(atom_ids)
        # Normal must always be acute with reference vector. This
        # ensures the normal is always the same, even if the molecule
        # is rotated.
        atom1, atom2 = self.get_atom_positions(atom_ids[:2])
        reference = atom1 - atom2
        normal = np.linalg.svd(pos - centroid)[-1][2, :]
        if (
            # vector_angle is NaN if cc_vector is [0, 0, 0].
            not np.allclose(reference, [0, 0, 0], atol=1e-5)
            and vector_angle(normal, reference) > np.pi/2
        ):
            normal *= -1
        return normal

    @classmethod
    def has_cached_mol(cls, identity_key):
        """
        ``True`` if molecule with `identity_key` is cached.

        Parameters
        ----------
        identity_key : :class:`object`
            The identity key of a molecule.

        Returns
        -------
        :class:`bool`
            ``True`` if a molecule with `identity_key` is cached.

        """

        return identity_key in cls._cache

    def get_position_matrix(self):
        """
        Return a matrix holding the atomic positions.

        Returns
        -------
        :class:`numpy.ndarray`
            The array has the shape ``(n, 3)``. Each row holds the
            x, y and z coordinates of an atom.

        """

        return np.array(self._position_matrix.T)

    def set_centroid(self, position, atom_ids=None):
        """
        Set the centroid to `position`.

        Parameters
        ----------
        position : :class:`numpy.ndarray`
            This array holds the position on which the centroid of the
            molecule is going to be placed.

        atom_ids : :class:`iterable` of :class:`int`
            The ids of atoms which should have their centroid set to
            `position`. If ``None`` then all atoms are used.

        Returns
        -------
        :class:`Molecule`
            The molecule.

        """

        centroid = self.get_centroid(atom_ids=atom_ids)
        self.apply_displacement(position-centroid)
        return self

    def set_position_matrix(self, position_matrix):
        """
        Set the coordinates to those in `position_matrix`.

        Parameters
        ----------
        position_matrix : :class:`numpy.ndarray`
            A position matrix of the molecule. The shape of the matrix
            is ``(n, 3)``.

        Returns
        -------
        :class:`Molecule`
            The molecule.

        """

        self._position_matrix = np.array(position_matrix.T)
        return self

    def dump(
        self,
        path,
        include_attrs=None,
        ignore_missing_attrs=False
    ):
        """
        Write a :class:`dict` representation to a file.

        Parameters
        ----------
        path : :class:`str`
            The full path to the file to which the  :class:`dict`
            should be written.

        include_attrs : :class:`list` of :class:`str`, optional
            The names of attributes of the molecule to be added to
            the representation. Each attribute is saved as a string
            using :func:`repr`.

        ignore_missing_attrs : :class:`bool`, optional
            If ``False`` and an attribute in `include_attrs` is not
            held by the :class:`Molecule`, an error will be raised.

        Returns
        -------
        None : :class:`NoneType`

        """

        with open(path, 'w') as f:
            d = self.to_dict(include_attrs, ignore_missing_attrs)
            json.dump(d, f, indent=4)

    @classmethod
    def load(cls, path, use_cache=False):
        """
        Initialize from a dump file.

        The :class:`Molecule` returned has the class specified in
        in the file, not :class:`Molecule`. You can use this if you
        don't know what class the instance in the loaded molecule is or
        should be.

        Parameters
        ----------
        path : :class:`str`
            The full path holding a dumped molecule.

        use_cache : :class:`bool`, optional
            If ``True``, a new instance will not be made if a cached
            and identical one already exists, the one which already
            exists will be returned. If ``True`` and a cached,
            identical instance does not yet exist the created one will
            be added to the cache.

        Returns
        -------
        :class:`Molecule`
            The molecule held in `path`.

        """

        with open(path, 'r') as f:
            mol_dict = json.load(f)

        return cls.init_from_dict(mol_dict, use_cache=use_cache)

    def _to_mdl_mol_block(self, atom_ids=None):
        """
        Return a V3000 mol block of the molecule.

        Parameters
        ----------
        atom_ids : :class:`iterable` of :class:`int`, optional
            The atom ids of atoms to write. If ``None`` then all atoms
            are written.

        Returns
        -------
        :class:`str`
            The V3000 mol block representing the molecule.

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))

        atom_lines = []
        # This set gets used by bonds.
        atoms = set()
        for i, atom_id in enumerate(atom_ids, 1):
            atoms.add(atom_id)

            x, y, z = self._position_matrix[:, atom_id]
            atom = self.atoms[atom_id]
            symbol = atom.__class__.__name__
            charge = atom.charge
            charge = f' CHG={charge}' if charge else ''
            atom_lines.append(
                'M  V30 {} {} {:.4f} {:.4f} {:.4f} 0{}\n'.format(
                    atom_id+1, symbol, x, y, z, charge
                )
            )
        atom_block = ''.join(atom_lines)
        num_atoms = i

        bond_lines = []
        for bond_idx, bond in enumerate(self.bonds):
            a1 = bond.atom1.id
            a2 = bond.atom2.id
            if a1 in atoms and a2 in atoms:
                # Keep bond ids if all bonds are getting written.
                if num_atoms == len(self.atoms):
                    bond_id = bond_idx
                else:
                    bond_id = len(bond_lines)

                bond_lines.append(
                    f'M  V30 {bond_id+1} '
                    f'{int(bond.order)} {a1+1} {a2+1}\n'
                )

        num_bonds = len(bond_lines)
        bond_block = ''.join(bond_lines)
        return (
            '\n'
            '     RDKit          3D\n'
            '\n'
            '  0  0  0  0  0  0  0  0  0  0999 V3000\n'
            'M  V30 BEGIN CTAB\n'
            f'M  V30 COUNTS {num_atoms} {num_bonds} 0 0 0\n'
            'M  V30 BEGIN ATOM\n'
            f'{atom_block}'
            'M  V30 END ATOM\n'
            'M  V30 BEGIN BOND\n'
            f'{bond_block}'
            'M  V30 END BOND\n'
            'M  V30 END CTAB\n'
            'M  END\n'
            '\n'
            '$$$$\n'
        )

    def to_rdkit_mol(self):
        """
        Return an :mod:`rdkit` representation.

        Returns
        -------
        :class:`rdkit.Mol`
            The molecule in :mod:`rdkit` format.

        """

        mol = rdkit.EditableMol(rdkit.Mol())
        for atom in self.atoms:
            rdkit_atom = rdkit.Atom(atom.atomic_number)
            rdkit_atom.SetFormalCharge(atom.charge)
            mol.AddAtom(rdkit_atom)

        for bond in self.bonds:
            mol.AddBond(
                beginAtomIdx=bond.atom1.id,
                endAtomIdx=bond.atom2.id,
                order=rdkit.BondType(bond.order)
            )

        mol = mol.GetMol()
        rdkit_conf = rdkit.Conformer(len(self.atoms))
        for atom_id, atom_coord in enumerate(self._position_matrix.T):
            rdkit_conf.SetAtomPosition(atom_id, atom_coord)
            mol.GetAtomWithIdx(atom_id).SetNoImplicit(True)
        mol.AddConformer(rdkit_conf)
        return mol

    def to_dict(self, include_attrs=None, ignore_missing_attrs=False):
        """
        Return a :class:`dict` representation.

        Parameters
        ----------
        include_attrs : :class:`list` of :class:`str`, optional
            The names of additional attributes of the molecule to be
            added to the :class:`dict`. Each attribute is saved as a
            string using :func:`repr`.

        ignore_missing_attrs : :class:`bool`, optional
            If ``False`` and an attribute in `include_attrs` is not
            held by the :class:`Molecule`, an error will be raised.

        Returns
        -------
        :class:`dict`
            A :class:`dict` representation of the molecule.

        Raises
        ------
        :class:`NotImplementedError`
            This is a virtual method, which needs to be implemented by
            a subclass.

        """

        raise NotImplementedError()

    def update_cache(self):
        """
        Update attributes of the cached molecule.

        If there is no identical molecule in the cache, then this
        molecule is added.

        When using multiprocessing, modified copies of the original
        molecules are created. In order to ensure that the cached
        molecules have their attributes updated to the values of the
        copies, this method must be run on the copies.

        Returns
        -------
        None : :class:`NoneType`

        """

        if self._identity_key in self.__class__._cache:
            d = dict(vars(self))
            self.__class__._cache[self._identity_key].__dict__ = d
        else:
            self.__class__._cache[self._identity_key] = self

    def update_from_rdkit_mol(self, mol):
        """
        Update the structure to match `mol`.

        Parameters
        ----------
        mol : :class:`rdkit.Mol`
            The :mod:`rdkit` molecule to use for the structure update.

        Returns
        -------
        :class:`.Molecule`
            The molecule.

        """

        pos_mat = mol.GetConformer().GetPositions()
        self.set_position_matrix(pos_mat)
        return self

    def update_from_file(self, path):
        """
        Update the structure from a file.

        Multiple file types are supported, namely:

        #. ``.mol``, ``.sdf`` - MDL V2000 and V3000 files
        #. ``.xyz`` - XYZ files
        #. ``.mae`` - Schrodinger Maestro files
        #. ``.coord`` - Turbomole files

        Parameters
        ----------
        path : :class:`str`
            The path to a molecular structure file holding updated
            coordinates for the :class:`.Molecule`.

        Returns
        -------
        :class:`.Molecule`
            The molecule.

        """

        update_fns = {
            '.mol': self._update_from_mol,
            '.sdf': self._update_from_mol,
            '.mae': self._update_from_mae,
            '.xyz': self._update_from_xyz,
            '.coord': self._update_from_turbomole
        }
        _, ext = os.path.splitext(path)
        update_fns[ext](path=path)
        return self

    def _update_from_mae(self, path):
        """
        Update the structure to match an ``.mae`` file.

        Parameters
        ----------
        path : :class:`str`
            The full path of the ``.mae`` file from which the structure
            should be updated.

        Returns
        -------
        None : :class:`NoneType`

        """

        self.update_from_rdkit_mol(mol_from_mae_file(path))

    def _update_from_mol(self, path):
        """
        Update the structure to match a ``.mol`` file.

        Parameters
        ----------
        path : :class:`str`
            The full path of the ``.mol`` file from which the structure
            should be updated.

        Returns
        -------
        None : :class:`NoneType`

        """

        mol = remake(
            rdkit.MolFromMolFile(
                molFileName=path,
                sanitize=False,
                removeHs=False
            )
        )
        self.update_from_rdkit_mol(mol=mol)

    def _update_from_xyz(self, path):
        """
        Update the structure to match an ``.xyz`` file.

        Parameters
        ----------
        path : :class:`str`
            The full path of the ``.mol`` file from which the structure
            should be updated.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`RuntimeError`
            If the number of atoms in the file does not match the
            number of atoms in the molecule or if atom elements in the
            file do not agree with the atom elements in the molecule.

        """

        with open(path, 'r') as f:
            atom_count, _, *content = f.readlines()

        # Check the atom count is correct.
        num_atoms = len(self.atoms)
        if int(atom_count) != num_atoms:
            raise RuntimeError(
                f'The number of atoms in the xyz file, {atom_count}, '
                'does not match the number of atoms in the '
                f'molecule, {num_atoms}.'
            )

        # Save all the coords in the file.
        new_coords = []
        for i, line in enumerate(content):
            element, *coords = line.split()
            if element.isnumeric():
                element = periodic_table[int(element)]

            if element != self.atoms[i].__class__.__name__:
                raise RuntimeError(
                    f'Atom {i} element does not match file.'
                )

            new_coords.append([float(i) for i in coords])

        # Check that the correct number of atom
        # lines was present in the file.
        if i+1 != num_atoms:
            raise RuntimeError(
                f'The number of atom lines in the xyz file, {i+1}, '
                'does not match the number of atoms in the '
                f'molecule, {num_atoms}.'
            )

        # Update the structure.
        new_coords = np.array(new_coords)
        self.set_position_matrix(new_coords)

    def _update_from_turbomole(self, path):
        """
        Update the structure from a Turbomole ``.coord`` file.

        Note that coordinates in ``.coord`` files are given in Bohr.

        Parameters
        ----------
        path : :class:`str`
            The full path of the ``.coord`` file from which the
            structure should be updated.

        Returns
        -------
        None : :class:`NoneType`

        Raises
        ------
        :class:`RuntimeError`
            If the number of atoms in the file does not match the
            number of atoms in the molecule or if atom elements in the
            file do not agree with the atom elements in the molecule.

        """
        bohr_to_ang = 0.5291772105638411

        with open(path, 'r') as f:
            _, *content, __ = f.readlines()

        # Check the atom count is correct.
        num_atoms = len(self.atoms)
        if len(content) != num_atoms:
            raise RuntimeError(
                'The number of atoms in the coord file, '
                f'{len(content)}, does not match the number of atoms '
                f'in the molecule, {num_atoms}.'
            )

        # Save all the coords in the file.
        new_coords = []
        for i, line in enumerate(content):
            *coords, element = line.split()
            if element.isnumeric():
                element = periodic_table[int(element)]

            if element != self.atoms[i].__class__.__name__:
                raise RuntimeError(
                    f'Atom {i} element does not match file.'
                )

            new_coords.append([float(i)*bohr_to_ang for i in coords])

        # Check that the correct number of atom
        # lines was present in the file.
        if i+1 != num_atoms:
            raise RuntimeError(
                'The number of atoms lines in the coord file, '
                f'{i+1}, does not match the number of atoms '
                f'in the molecule, {num_atoms}.'
            )

        # Update the structure.
        new_coords = np.array(new_coords)
        self.set_position_matrix(new_coords)

    def write(self, path, atom_ids=None):
        """
        Write the structure to a file.

        This function will write the format based on the extension
        of `path`.

        #. ``.mol``, ``.sdf`` - MDL V3000 MOL file
        #. ``.xyz`` - XYZ file
        #. ``.pdb`` - PDB file

        Parameters
        ----------
        path : :class:`str`
            The `path` to which the molecule should be written.

        atom_ids : :class:`iterable` of :class:`int`, optional
            The atom ids of atoms to write. If ``None`` then all atoms
            are written.

        Returns
        -------
        :class:`.Molecule`
            The molecule.

        """

        write_funcs = {
            '.mol': self._write_mdl_mol_file,
            '.sdf': self._write_mdl_mol_file,
            '.xyz': self._write_xyz_file,
            '.pdb': self._write_pdb_file
        }

        _, ext = os.path.splitext(path)
        write_func = write_funcs[ext]
        write_func(path, atom_ids)
        return self

    def _write_mdl_mol_file(self, path, atom_ids):
        """
        Write to a V3000 ``.mol`` file.

        This function should not be used directly, only via
        :meth:`write`.

        Parameters
        ----------
        path : :class:`str`
            The full path to the file being written.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. If ``None`` then all atoms
            are written.

        Returns
        -------
        None : :class:`NoneType`

        """

        with open(path, 'w') as f:
            f.write(self._to_mdl_mol_block(atom_ids))

    def _write_xyz_file(self, path, atom_ids):
        """
        Write to a ``.xyz`` file.

        This function should not be used directly, only via
        :meth:`write`.

        Parameters
        ----------
        path : :class:`str`
            The full path to the file being written.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. If ``None`` then all atoms
            are written.

        Returns
        -------
        None : :class:`NoneType`

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))

        content = [0]
        for i, atom_id in enumerate(atom_ids, 1):
            x, y, z = self._position_matrix[:, atom_id]
            symbol = self.atoms[atom_id].__class__.__name__
            content.append(f'{symbol} {x:f} {y:f} {z:f}\n')
        # Set first line to the atom_count.
        content[0] = f'{i}\n\n'

        with open(path, 'w') as xyz:
            xyz.write(''.join(content))

    def _write_pdb_file(self, path, atom_ids):
        """
        Write to a ``.pdb`` file.

        This function should not be used directly, only via
        :meth:`write`.

        Parameters
        ----------
        path : :class:`str`
            The full path to the file being written.

        atom_ids : :class:`iterable` of :class:`int`
            The atom ids of atoms to write. If ``None`` then all atoms
            are written.

        Returns
        -------
        None : :class:`NoneType`

        """

        if atom_ids is None:
            atom_ids = range(len(self.atoms))

        lines = []
        atom_counts = {}
        hetatm = 'HETATM'
        alt_loc = ''
        res_name = 'UNL'
        chain_id = ''
        res_seq = '1'
        i_code = ''
        occupancy = '1.00'
        temp_factor = '0.00'

        coords = self._position_matrix
        # This set will be used by bonds.
        atoms = set()
        for atom in atom_ids:
            atoms.add(atom)

            serial = atom+1
            element = self.atoms[atom].__class__.__name__
            atom_counts[element] = atom_counts.get(element, 0) + 1
            name = f'{element}{atom_counts[element]}'
            # Make sure the coords are no more than 8 columns wide
            # each.
            x, y, z = (f'{i}'[:8] for i in coords[:, atom])
            lines.append(
                f'{hetatm:<6}{serial:>5} {name:<4}'
                f'{alt_loc:<1}{res_name:<3} {chain_id:<1}'
                f'{res_seq:>4}{i_code:<1}   '
                f'{x:>8}{y:>8}{z:>8}'
                f'{occupancy:>6}{temp_factor:>6}          '
                f'{element:>2}{self.atoms[atom].charge:>2}\n'
            )

        conect = 'CONECT'
        for bond in self.bonds:
            a1 = bond.atom1.id
            a2 = bond.atom2.id
            if a1 in atoms and a2 in atoms:
                lines.append(
                    f'{conect:<6}{a1+1:>5}{a2+1:>5}               \n'
                )

        lines.append('END\n')
        with open(path, 'w') as f:
            f.write(''.join(lines))



"""
Constructed Molecule
====================

"""

import logging
import numpy as np
from collections import Counter

from .. import elements, bonds, topology_graphs
from .molecule import Molecule
from ..functional_groups import FunctionalGroup, fg_types

logger = logging.getLogger(__name__)


class ConstructionError(Exception):
    ...


class ConstructedMolecule(Molecule):
    """
    Represents constructed molecules.

    A :class:`ConstructedMolecule` requires at least 2 basic pieces of
    information: which building block molecules are used to construct
    the molecule and what the :class:`.TopologyGraph` of the
    constructed molecule is. The construction of the molecular
    structure is performed by :meth:`.TopologyGraph.construct`. This
    method does not have to be called explicitly by the user, it will
    be called automatically during initialization.

    The building block molecules used for construction can be either
    :class:`.BuildingBlock` instances or other
    :class:`.ConstructedMolecule` instances, or a combination both.

    Each :class:`.TopologyGraph` subclass may add additional attributes
    to the :class:`ConstructedMolecule`, which will be described within
    its documentation.

    Attributes
    ----------
    atoms : :class:`tuple` of :class:`.Atom`
        The atoms of the molecule. Each :class:`.Atom`
        instance is guaranteed to have two attributes. The
        first is :attr:`building_block`, which holds the building
        block :class:`.Molecule` from which that
        :class:`.Atom` came. If the :class:`.Atom` did not come from a
        building block, but was added by a reaction, the value
        of this attribute will be ``None``.

        The second attribute is :attr:`building_block_id`.
        This will be the same value on all atoms that came from the
        building block. Note that if a building block is used multiple
        times during construction, the :attr:`building_block_id` will
        be different for each time it is used.

    bonds : :class:`tuple` of :class:`.Bond`
        The bonds of the molecule.

    building_block_vertices : :class:`dict`
        Maps the :class:`.Molecule` instances used for construction,
        which can be either :class:`.BuildingBlock` or
        :class:`.ConstructedMolecule`, to the
        :class:`~.topologies.base.Vertex` objects they are placed on
        during construction. The :class:`dict` has the form

        .. code-block:: python

            building_block_vertices = {
                BuildingBlock(...): [Vertex(...), Vertex(...)],
                BuildingBlock(...): [
                    Vertex(...),
                    Vertex(...),
                    Vertex(...),
                ]
                ConstructedMolecule(...): [Vertex(...)]
            }

    building_block_counter : :class:`collections.Counter`
        A counter keeping track of how many times each building block
        molecule appears in the :class:`ConstructedMolecule`.

    topology_graph : :class:`.TopologyGraph`
        Defines the topology graph of :class:`ConstructedMolecule` and
        is responsible for constructing it.

    construction_bonds : :class:`tuple` of :class:`.Bond`
        Holds the bonds in :attr:`bonds`, which were added by the
        construction process.

    func_groups : :class:`tuple` of :class:`.FunctionalGroup`
        The remnants of building block functional groups present in the
        molecule. They track which atoms belonged to functional groups
        in the building block molecules. The id of each
        :class:`.FunctionalGroup` should match its index in
        :attr:`func_groups`.

    Examples
    --------
    *Initialization*

    A :class:`ConstructedMolecule` can be created from a set of
    building blocks and a :class:`.TopologyGraph`

    .. code-block:: python

        import stk

        bb1 = stk.BuildingBlock('NCCCN', ['amine'])
        bb2 = stk.BuildingBlock('O=CC(C=O)CC=O', ['aldehyde'])
        tetrahedron = stk.cage.FourPlusSix()
        cage1 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=tetrahedron
        )

    A :class:`ConstructedMolecule` can be used to construct other
    :class:`ConstructedMolecule` instances

    .. code-block:: python

        benzene = stk.BuildingBlock('c1ccccc1')
        cage_complex = stk.ConstructedMolecule(
            building_blocks=[cage1, benzene],
            topology_graph=stk.host_guest.Complex()
        )

    During initialization it is possible to force building blocks to
    be placed on specific :attr:`~.TopologyGraph.vertices` of the
    :class:`.TopologyGraph`

    .. code-block:: python

        bb3 = stk.BuildingBlock('NCOCN', ['amine'])
        bb4 = stk.BuildingBlock('NCOCCCOCN', ['amine'])
        cage2 = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2, bb3, bb4],
            topology_graph=tetrahedron
            building_block_vertices={
                bb1: tetrahedron.vertices[4:6]
                bb3: tetrahedron.vertices[6:7]
                bb4: tetrahedron.vertices[7:8]
            }
        )

    *Building blocks with the wrong number of functional groups.*

    If the building block has too many functional groups, you can
    remove some in order to use it

    .. code-block:: python

        chain = stk.polymer.Linear('AB', [0, 0],  3)

        # This won't work, bb2 has 3 functional groups but 2 are needed
        # for monomers in a linear polymer chain.
        failed = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=chain
        )

        # Remove one of the functional groups and you will be able to
        # construct the chain.
        bb2.func_groups = (bb2.func_groups[0], bb2.func_groups[2])
        failed = stk.ConstructedMolecule(
            building_blocks=[bb1, bb2],
            topology_graph=chain
        )

    """

    @classmethod
    def _construct(
        cls,
        building_blocks,
        topology_graph,
        building_block_vertices=None,
        use_cache=False
    ):
        if building_block_vertices is None:
            building_block_vertices = (
                topology_graph.assign_building_blocks_to_vertices(
                    building_blocks=building_blocks
                )
            )
        identity_key = cls._get_identity_key_from_components(
            building_blocks=building_blocks,
            topology_graph=topology_graph,
            building_block_vertices=building_block_vertices
        )
        if use_cache and identity_key in cls._cache:
            return cls._cache[identity_key]

        obj = cls._init_from_components(
            building_blocks=building_blocks,
            topology_graph=topology_graph,
            building_block_vertices=building_block_vertices,
            identity_key=identity_key
        )
        if use_cache:
            cls._cache[identity_key] = obj
        return obj

    def __init__(
        self,
        building_blocks,
        topology_graph,
        building_block_vertices=None,
        use_cache=False
    ):
        """
        Initialize a :class:`ConstructedMolecule`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        topology_graph : :class:`.TopologyGraph`
            Defines the topology graph of the
            :class:`ConstructedMolecule` and constructs it.

        building_block_vertices : :class:`dict`, optional
            Maps the :class:`.Molecule` in  `building_blocks` to the
            :class:`~.topologies.base.Vertex` instances in
            `topology_graph` it is placed on. Each
            :class:`.BuildingBlock` and :class:`ConstructedMolecule`
            can be mapped to multiple :class:`~.topologies.base.Vertex`
            objects. See the examples section in the
            :class:`.ConstructedMolecule` class docstring to help
            understand how this parameter is used. If ``None``,
            building block molecules will be assigned to vertices at
            random.

        use_cache : :class:`bool`, optional
            If ``True``, a new :class:`.ConstructedMolecule` will
            not be made if a cached and identical one already exists,
            the one which already exists will be returned. If ``True``
            and a cached, identical :class:`ConstructedMolecule` does
            not yet exist the created one will be added to the cache.

        """

        # This method does not get called, See _construct().
        raise RuntimeError('This method should not be getting called.')

    @classmethod
    def _init_from_components(
        cls,
        building_blocks,
        topology_graph,
        building_block_vertices,
        identity_key
    ):
        """
        Initialize a :class:`ConstructedMolecule`.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        topology_graph : :class:`.TopologyGraph`
            Defines the topology graph of the
            :class:`ConstructedMolecule` and constructs it.

        building_block_vertices : :class:`dict`
            Maps the :class:`.Molecule` in  `building_blocks` to the
            :class:`~.topologies.base.Vertex` in `topology_graph`.
            Each :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` can be mapped to multiple
            :class:`~.topologies.base.Vertex` objects. See the
            examples section in the :class:`.ConstructedMolecule`
            class docstring to help understand how this parameter
            is used. If ``None``, building block molecules will be
            assigned to vertices at random.

        identity_key : :class:`tuple`
            The identity key of the molecule. The identity key wil be
            equal for two molecules which ``stk`` sees as identical.

        Returns
        -------
        :class:`.ConstructedMolecule`
            The molecule.

        """

        obj = cls.__new__(cls)
        obj._identity_key = identity_key
        obj.building_block_vertices = building_block_vertices
        obj.topology_graph = topology_graph
        obj.atoms = []
        obj.bonds = []
        obj.construction_bonds = []
        obj.func_groups = []
        obj.building_block_counter = Counter()
        # A (3, n) numpy.ndarray holding the position of every atom in
        # the molecule.
        obj._position_matrix = []

        try:
            topology_graph.construct(obj)

        except Exception as ex:
            errormsg = (
                'Construction failure.\n'
                '\n'
                'topology_graph\n'
                '--------------\n'
                f'{topology_graph}\n'
                '\n'
                'building blocks\n'
                '---------------\n'
            )

            bb_blocks = []
            for i, bb in enumerate(building_blocks):
                bb_blocks.append(
                    f'{bb}\n\n'
                    'MDL MOL BLOCK\n'
                    '-------------\n'
                    f'{bb._to_mdl_mol_block()}'
                )

            errormsg += '\n'.join(bb_blocks)
            raise ConstructionError(errormsg) from ex

        obj.atoms = tuple(obj.atoms)
        obj.bonds = tuple(obj.bonds)
        obj.construction_bonds = tuple(obj.construction_bonds)
        obj.func_groups = tuple(obj.func_groups)

        # Ensure that functional group ids are set correctly.
        for id_, func_group in enumerate(obj.func_groups):
            func_group.id = id_

        return obj

    def get_building_blocks(self):
        """
        Yield the building blocks.

        Yields
        ------
        :class:`.Molecule`
            A building block of the :class:`ConstructedMolecule`.

        """

        yield from self.building_block_vertices.keys()

    def to_dict(self, include_attrs=None, ignore_missing_attrs=False):
        """
        Return a :class:`dict` representation.

        Parameters
        ----------
        include_attrs : :class:`list` of :class:`str`, optional
            The names of additional attributes of the molecule to be
            added to the :class:`dict`. Each attribute is saved as a
            string using :func:`repr`. These attributes are also
            passed down recursively to the building block molecules.

        ignore_missing_attrs : :class:`bool`, optional
            If ``False`` and an attribute in `include_attrs` is not
            held by the :class:`ConstructedMolecule`, an error will be
            raised.

        Returns
        -------
        :class:`dict`
            A :class:`dict` which represents the molecule.

        """

        if include_attrs is None:
            include_attrs = []

        bb_counter = []
        building_blocks = []
        building_block_vertices = {}
        bb_index = {}
        for i, bb in enumerate(self.building_block_vertices):
            bb_index[bb] = i
            bb_counter.append(self.building_block_counter[bb])
            building_blocks.append(bb.to_dict(include_attrs, True))
            building_block_vertices[i] = [
                self.topology_graph.vertices.index(v)
                for v in self.building_block_vertices[bb]
            ]

        # For atoms, bond sand func groups, need to change all atom
        # references with the id of the atom. This is because when
        # these objects a reconstructed you want to create pointers
        # back to the atoms found in the atoms attribute.
        atoms = []
        for atom in self.atoms:
            clone = atom.clone()
            # This attribute is a pointer that will need to get
            # re-created by load().
            clone.building_block = bb_index[clone.building_block]
            atoms.append(clone)

        bonds = []
        bond_indices = {}
        for i, bond in enumerate(self.bonds):
            clone = bond.clone()
            clone.atom1 = clone.atom1.id
            clone.atom2 = clone.atom2.id
            bonds.append(clone)
            bond_indices[bond] = i

        construction_bonds = [
            bond_indices[bond] for bond in self.construction_bonds
        ]

        func_groups = []
        for fg in self.func_groups:
            clone = fg.clone()
            clone.atoms = tuple(clone.get_atom_ids())
            clone.bonders = tuple(clone.get_bonder_ids())
            clone.deleters = tuple(clone.get_deleter_ids())
            clone.fg_type = f'{clone.fg_type.name!r}'
            func_groups.append(clone)

        d = {
            'building_blocks': building_blocks,
            'building_block_vertices': building_block_vertices,
            'building_block_counter': bb_counter,
            'construction_bonds': construction_bonds,
            'class': self.__class__.__name__,
            'position_matrix': self.get_position_matrix().tolist(),
            'topology_graph': repr(self.topology_graph),
            'func_groups': repr(tuple(func_groups)),
            'atoms': repr(tuple(atoms)),
            'bonds': repr(tuple(bonds)),
            'identity_key': repr(self._identity_key),
        }

        if ignore_missing_attrs:
            d.update({
                attr: repr(getattr(self, attr))
                for attr in include_attrs
                if hasattr(self, attr)
            })
        else:
            d.update({
                attr: repr(getattr(self, attr))
                for attr in include_attrs
            })

        return d

    @classmethod
    def _init_from_dict(cls, mol_dict, use_cache):
        """
        Intialize from a :class:`dict` representation.

        Parameters
        ----------
        mol_dict : :class:`dict`
            A :class:`dict` representation of a molecule generated
            by :meth:`to_dict`.

        use_cache : :class:`bool`
            If ``True``, a new instance will not be made if a cached
            and identical one already exists, the one which already
            exists will be returned. If ``True`` and a cached,
            identical instance does not yet exist the created one will
            be added to the cache.

        Returns
        -------
        :class:`ConstructedMolecule`
            The molecule described by `mol_dict`.

        """

        d = dict(mol_dict)
        identity_key = eval(d.pop('identity_key'))
        if use_cache and identity_key in cls._cache:
            return cls._cache[identity_key]

        tops = vars(topology_graphs)
        topology_graph = eval(d.pop('topology_graph'), tops)
        bbs = [
            Molecule.init_from_dict(bb_dict, use_cache)
            for bb_dict in d.pop('building_blocks')
        ]

        obj = cls.__new__(cls)
        obj._identity_key = identity_key
        obj.building_block_counter = Counter()
        obj.building_block_vertices = {}

        counter = d.pop('building_block_counter')
        vertices = d.pop('building_block_vertices')
        for i, bb in enumerate(bbs):
            obj.building_block_counter[bb] = counter[i]
            obj.building_block_vertices[bb] = [
                topology_graph.vertices[i] for i in vertices[str(i)]
            ]

        obj.topology_graph = topology_graph
        obj._position_matrix = np.array(d.pop('position_matrix')).T

        obj.atoms = eval(d.pop('atoms'), vars(elements))
        for atom in obj.atoms:
            atom.building_block = bbs[atom.building_block]

        obj.bonds = eval(d.pop('bonds'), vars(bonds))
        for bond in obj.bonds:
            bond.atom1 = obj.atoms[bond.atom1]
            bond.atom2 = obj.atoms[bond.atom2]

        obj.construction_bonds = [
            obj.bonds[i] for i in d.pop('construction_bonds')
        ]

        g = {'FunctionalGroup': FunctionalGroup}
        obj.func_groups = tuple(eval(d.pop('func_groups'), g))
        for func_group in obj.func_groups:
            func_group.atoms = tuple(
                obj.atoms[i] for i in func_group.atoms
            )
            func_group.bonders = tuple(
                obj.atoms[i] for i in func_group.bonders
            )
            func_group.deleters = tuple(
                obj.atoms[i] for i in func_group.deleters
            )
            func_group.fg_type = fg_types[func_group.fg_type]

        d.pop('class')
        for attr, val in d.items():
            setattr(obj, attr, eval(val))

        if use_cache:
            cls._cache[identity_key] = obj
        return obj

    @classmethod
    def _get_identity_key_from_components(
        cls,
        building_blocks,
        topology_graph,
        building_block_vertices,
    ):
        """
        Return the identity key.

        Parameters
        ----------
        building_blocks : :class:`list` of :class:`.Molecule`
            The :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` instances which
            represent the building block molecules used for
            construction. Only one instance is present per building
            block molecule, even if multiples of that building block
            join up to form the :class:`ConstructedMolecule`.

        topology_graph : :class:`.TopologyGraph`
            Defines the topology graph of the
            :class:`ConstructedMolecule` and constructs it.

        building_block_vertices : :class:`dict`
            Maps the :class:`.Molecule` in  `building_blocks` to the
            :class:`~.topologies.base.Vertex` in `topology_graph`.
            Each :class:`.BuildingBlock` and
            :class:`ConstructedMolecule` can be mapped to multiple
            :class:`~.topologies.base.Vertex` objects. See the
            examples section in the :class:`.ConstructedMolecule`
            class docstring to help understand how this parameter
            is used. If ``None``, building block molecules will be
            assigned to vertices at random.

        Returns
        -------
        :class:`tuple`
            The identity key.

        """

        building_block_types = sorted(
            {bb.__class__ for bb in building_blocks},
            key=lambda cls: cls.__name__
        )
        bb_keys = cls._get_sorted_building_block_keys(
            building_blocks=building_blocks,
            building_block_types=building_block_types
        )
        vertices = cls._get_sorted_building_block_vertices(
            building_blocks=building_blocks,
            building_block_vertices=building_block_vertices,
            building_block_types=building_block_types
        )
        return tuple(bb_keys), repr(topology_graph), tuple(vertices)

    @classmethod
    def _get_sorted_building_block_keys(
        cls,
        building_blocks,
        building_block_types
    ):
        for bb_type in building_block_types:
            bbs = filter(
                lambda bb: bb.__class__ is bb_type,
                building_blocks
            )
            yield from sorted(x.get_identity_key() for x in bbs)

    @staticmethod
    def _get_building_block_vertices_of_type(
        building_block_vertices,
        building_block_type
    ):

        bbs = filter(
            lambda bb: bb.__class__ is building_block_type,
            building_block_vertices
        )
        for bb in bbs:
            vertices = tuple(sorted(
                v.id for v in building_block_vertices[bb]
            ))
            yield bb.get_identity_key(), vertices

    @classmethod
    def _get_sorted_building_block_vertices(
        cls,
        building_blocks,
        building_block_vertices,
        building_block_types
    ):
        # The identity key needs to account for which building block
        # is placed on which vertex. This means that a tuple matching
        # the identity key of each building block with each vertex
        # needs to be included. This tuple needs to be sorted so that
        # the identity key of a constructed molecule is always the
        # same. However, the constructed molecule can be constructed
        # with both BuildingBlocks and other ConstructedMolecules.
        # They each have differnt types as identity keys which means
        # they connot be sorted together. As a result, first
        # sort the identity keys of all buliding blocks of the same
        # type and then join these sorted tuples togther.
        # The sorted tuples are joined together in order based on the
        # names of the classes.
        for building_block_type in building_block_types:
            yield from sorted(
                cls._get_building_block_vertices_of_type(
                    building_block_vertices=building_block_vertices,
                    building_block_type=building_block_type
                )
            )

    def __str__(self):
        return (
            f'{self.__class__.__name__}'
            f'(building_blocks={list(self.get_building_blocks())}, '
            f'topology_graph={self.topology_graph!r})'
        )

    def __repr__(self):
        return str(self)
